# -*- coding: utf-8 -*-
"""
lories._core._connectors
~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import Collection, TypeAlias, TypeVar

import pandas as pd
from lories._core._channels import Channels
from lories._core._registrator import _Registrator, _RegistratorContext
from lories._core._resources import Resources


class ConnectType(Enum):
    NONE = "NONE"
    AUTO = "AUTO"

    @classmethod
    def get(cls, value: str | bool) -> ConnectType:
        if isinstance(value, str):
            value = value.lower()
            if value.upper() in ["auto", "true"]:
                return ConnectType.AUTO
            if value.upper() == ["none", "false"]:
                return ConnectType.NONE
        if isinstance(value, bool):
            if value:
                return ConnectType.AUTO
            else:
                return ConnectType.NONE
        raise ValueError("Unknown ConnectType: " + str(value))

    def __str__(self):
        return str(self.value)


class _Connector(_Registrator):
    INCLUDES: Collection[str] = ()
    TYPE: str = "connector"

    @property
    @abstractmethod
    def resources(self) -> Resources: ...

    @property
    @abstractmethod
    def channels(self) -> Channels: ...

    @abstractmethod
    def is_connected(self) -> bool: ...

    def connect(self, resources: Resources) -> None:
        pass

    def disconnect(self) -> None:
        pass

    @abstractmethod
    def read(self, resources: Resources) -> pd.DataFrame: ...

    @abstractmethod
    def write(self, data: pd.DataFrame) -> None: ...


Connector = TypeVar("Connector", bound=_Connector)


# noinspection PyAbstractClass
class _ConnectorContext(_RegistratorContext[Connector]):
    TYPE: str = "connectors"


ConnectorContext = TypeVar(
    name="ConnectorContext",
    bound=_ConnectorContext,
)
Connectors: TypeAlias = ConnectorContext



# -*- coding: utf-8 -*-
"""
lories.core.configs.toml
~~~~~~~~~~~~~~~~~~~~~~~~


"""

import re
from typing import Any, Mapping

try:
    import tomllib as toml
except ModuleNotFoundError:
    import tomli as toml


def load_toml(conf_path: str) -> Mapping[str, Any]:
    with open(conf_path, mode="r") as conf_file:
        conf_string = conf_file.read()
        conf_string = re.sub(r"^;+", "#", conf_string, flags=re.MULTILINE)

        return toml.loads(conf_string)



# -*- coding: utf-8 -*-
"""
lories.components.tariff.static
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Optional

import pandas as pd
from lories.components import ComponentError
from lories.components.tariff import Tariff, register_tariff_type
from lories.typing import Configurations, Timestamp


# noinspection SpellCheckingInspection
@register_tariff_type("static")
class StaticTariff(Tariff):
    _price_import: float
    _price_export: float

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)
        self._price_import = configs.get_float("import")
        self._price_export = configs.get_float("export", default=0)

    def get(
        self,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        freq: str = "15min",
        **kwargs,
    ) -> pd.DataFrame:
        if any(t is None for t in [start, end]):
            raise ComponentError(self, "Unable to generate static tariff for incomplete or missing time range")
        return pd.DataFrame(
            index=pd.date_range(start=start, end=end, freq=freq),
            data={
                Tariff.PRICE_IMPORT: self._price_import,
                Tariff.PRICE_EXPORT: self._price_export,
            },
        )



# -*- coding: utf-8 -*-
"""
lories.core.configs.directories
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import os
from collections.abc import Mapping
from pathlib import Path, PosixPath, WindowsPath
from typing import Dict, Optional

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import LiteralString

except ImportError:
    from typing_extensions import LiteralString


class Directories:
    TYPE = "directories"

    LIB = "lib_dir"
    LOG = "log_dir"
    TMP = "tmp_dir"
    DATA = "data_dir"
    CONF = "conf_dir"

    KEYS = [LIB, LOG, TMP, DATA, CONF]

    def __init__(
        self,
        lib_dir: str = None,
        log_dir: str = None,
        tmp_dir: str = None,
        data_dir: str = None,
        conf_dir: str = None,
    ):
        self.lib = lib_dir
        self.log = log_dir
        self.tmp = tmp_dir
        self.data = data_dir
        self.conf = conf_dir

    # noinspection PyProtectedMember
    def __repr__(self) -> str:
        attrs = ["conf", "data", "tmp", "log", "lib"]
        return f"{type(self).__name__}({', '.join(f'{attr}={getattr(self, attr)._dir}' for attr in attrs)})"

    def __str__(self) -> str:
        attrs = ["conf", "data", "tmp", "log", "lib"]
        return f"[{self.TYPE}]\n" + "\n".join(f'{attr} = "{str(getattr(self, attr))}"' for attr in attrs)

    def __copy__(self) -> Directories:
        return self.copy()

    def __deepcopy__(self, memo) -> Directories:
        return self.__copy__()

    # noinspection PyProtectedMember
    def to_dict(self) -> Dict[str, Optional[str]]:
        dirs = {
            self.LIB: self._lib._dir,
            self.LOG: self._log._dir,
            self.TMP: self._tmp._dir,
            self.DATA: self._data._dir,
            self.CONF: self._conf._dir,
        }
        return dirs

    @property
    def lib(self) -> Directory:
        return self._lib

    # noinspection PyShadowingBuiltins
    @lib.setter
    def lib(self, dir: str | Directory) -> None:
        if isinstance(dir, Directory):
            dir = str(dir)
        self._lib = Directory(dir, default="lib")

    @property
    def log(self) -> Directory:
        return self._log

    # noinspection PyShadowingBuiltins
    @log.setter
    def log(self, dir: str | Directory) -> None:
        if isinstance(dir, Directory):
            dir = str(dir)
        self._log = Directory(dir, default="log")

    @property
    def tmp(self) -> Directory:
        return self._tmp

    # noinspection PyShadowingBuiltins
    @tmp.setter
    def tmp(self, dir: str | Directory) -> None:
        if isinstance(dir, Directory):
            dir = str(dir)
        self._tmp = Directory(dir, default="tmp")

    @property
    def data(self) -> Directory:
        return self._data

    # noinspection PyShadowingBuiltins
    @data.setter
    def data(self, dir: str | Directory) -> None:
        if isinstance(dir, Directory):
            dir = str(dir)
        self._data = Directory(dir, default="data")

    @property
    def conf(self) -> Directory:
        return self._conf

    # noinspection PyShadowingBuiltins
    @conf.setter
    def conf(self, dir: str | Path) -> None:
        if isinstance(dir, Path):
            dir = str(dir)
        self._conf = Directory(dir, default="conf", base=self.data)

    # noinspection PyProtectedMember
    def copy(self) -> Directories:
        return Directories(
            self._lib._dir,
            self._log._dir,
            self._tmp._dir,
            self._data._dir,
            self._conf._dir,
        )

    # noinspection PyProtectedMember, PyShadowingBuiltins
    def update(self, configs: Mapping[str, str]) -> None:
        for key in ["lib", "log", "tmp", "data"]:
            dir = configs.get(f"{key}_dir", None)
            if dir is not None:
                setattr(self, f"{key}", dir)
        conf_dir = configs.get("conf_dir", None)
        if conf_dir is not None:
            self.conf = conf_dir


class Directory(Path):
    _dir: Optional[Path] = None
    _base: Path
    default: str

    # noinspection PyShadowingBuiltins, PyTypeChecker
    def __new__(cls, *dirs: str, base: Optional[str | Directory] = None, default: Optional[str] = None):
        cls = WindowsDirectory if os.name == "nt" else PosixDirectory
        base = Directory.__parse_base(base)
        dir = Directory.__parse_dir(base, *dirs, default=default)
        return super().__new__(cls, Directory.__parse(base, dir, default), base=base, default=default)

    def __init__(self, *dirs: str, base: Optional[str | Directory] = None, default: Optional[str] = None):
        self.default = default
        self._base = Directory.__parse_base(base)
        self._dir = Directory.__parse_dir(self._base, *dirs, default=default)
        try:
            super().__init__(Directory.__parse(self._base, self._dir, default=default))

        except TypeError:
            # FixMe: The mro appears to be called incorrectly for python Versions < 3.12.
            # ToDo: Remove this catch, as older versions proceed to be deprecated.
            super().__init__()

    @staticmethod
    def __parse_base(base: Optional[LiteralString | str | Path]) -> Path:
        if base is None or (isinstance(base, Directory) and base.is_default()):
            base = Path.cwd()
        elif isinstance(base, Directory) or not isinstance(base, Path):
            base = Path(base)
        # FIXME: Remove this once Python >= 3.9 is a requirement
        # if base.is_relative_to("~"):
        if str(base).startswith("~"):
            base = base.expanduser()
        if not base.is_absolute():
            base = base.absolute()
        return base

    # noinspection PyShadowingBuiltins
    @staticmethod
    def __parse_dir(base: Path, *dirs: Optional[str], default: Optional[str] = None) -> Path:
        dir = Path(*dirs) if not any(d is None for d in dirs) else None
        if dir is not None:
            # FIXME: Remove this once Python >= 3.9 is a requirement
            # if dir.is_relative_to(base):
            if str(dir).startswith(str(base)):
                dir = dir.relative_to(base)
            # if dir.is_relative_to("~"):
            if str(dir).startswith("~"):
                dir = dir.expanduser()
        if str(dir) == default:
            dir = None
        return dir

    @staticmethod
    def __parse(base: Path, path: Optional[Path], default: Optional[str] = None) -> Path:
        if path is None:
            path = Path(default)
        if path is not None and not os.path.isabs(path):
            path = base.joinpath(path)
        return path

    def is_default(self) -> bool:
        if self.default is None:
            return False
        return self._dir is None or str(self) == os.path.join(self._base, self.default)

    # noinspection SpellCheckingInspection
    def joinpath(self, *paths: LiteralString | str | Path) -> Directory:
        _dir = self.__parse(self._base, self._dir, self.default)
        return Directory(_dir.joinpath(*paths), default=self.default, base=self._base)

    def relative_to(self, path: LiteralString | str | Path, *_, walk_up=False) -> Path:
        return self.__parse(self._base, self._dir, self.default).relative_to(path)


class PosixDirectory(Directory, PosixPath):
    pass


class WindowsDirectory(Directory, WindowsPath):
    pass



# -*- coding: utf-8 -*-
"""
lories.application.view.login.page
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Optional

import dash_bootstrap_components as dbc
from dash import Dash, Input, Output, State, dcc, html
from dash_auth import public_callback
from flask_login import LoginManager

from lories.application import Application
from lories.application.view.login import Authentication, AuthenticationException
from lories.application.view.pages import Page, PageLayout
from lories.core import Configurations


# noinspection SpellCheckingInspection
class LoginPage(Page):
    __manager: LoginManager
    __auth: Authentication

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        app: Dash,
        context: Application,
        configs: Configurations,
    ) -> None:
        super().__init__(id=f"{context.id}-login", key="login", name="Login")
        self.__auth = Authentication(app, configs, self.path)
        self.__manager = LoginManager(app.server)
        self.__manager.login_view = self.path
        self.__manager.user_loader(self.__auth.get_user)
        self.__manager.request_loader(self.__auth.request_user)

    def create_layout(self, layout: PageLayout) -> None:
        layout.container.class_name = "card-container auth-container"

        username_id = f"{self.id}-username"
        username_input = html.Div(
            [
                dbc.Label("Username", html_for=username_id),
                dbc.Input(type="username", id=username_id, placeholder="Username"),
            ],
            className="mb-3",
        )

        password_id = f"{self.id}-password"
        password_input = html.Div(
            [
                dbc.Label("Password", html_for=password_id),
                dbc.Input(type="password", id=password_id, placeholder="Password"),
            ],
            className="mb-3",
        )

        notice_id = f"{self.id}-notice"
        login_id = f"{self.id}-login"
        login_button = dbc.Button("Log in", id=login_id, n_clicks=0, color="primary", style={"width": "100%"})
        layout.append(
            dbc.Row(
                dbc.Col(
                    [
                        dbc.Row(dbc.Card(dbc.CardBody(dbc.Form([username_input, password_input, login_button])))),
                        dbc.Row(id=notice_id, style={"min-height": "60px", "padding-top": "12px"}),
                    ],
                    width="auto",
                ),
                align="center",
                justify="center",
                style={"height": "100%"},
            )
        )

        @public_callback(
            Output(notice_id, "children"),
            Input(login_id, "n_clicks"),
            [
                State(username_id, "value"),
                State(password_id, "value"),
            ],
            prevent_initial_call=True,
        )
        def update_output(_, username, password) -> Optional[dbc.Alert] | dcc.Location:
            try:
                if username == "" or username is None:
                    failure = "Username is empty"
                elif password == "" or password is None:
                    failure = "Password is empty"
                elif not self.__auth.login_user(username, password):
                    failure = "Unknown Error"
                else:
                    return dcc.Location(id=f"{self.id}-success", pathname="/")

            except AuthenticationException as e:
                failure = e.reason

            return dbc.Alert(
                children=[
                    html.H5("Log in failed", className="alert-heading"),
                    html.P(failure),
                ],
                color="danger",
            )



# -*- coding: utf-8 -*-
"""
lories.data.converters
~~~~~~~~~~~~~~~~~~~~~~


"""

from .errors import ConversionError  # noqa: F401

from .converter import Converter  # noqa: F401

from . import access  # noqa: F401
from .access import ConverterAccess  # noqa: F401

from . import context  # noqa: F401
from .context import (  # noqa: F401
    ConverterContext,
    register_converter_type,
    registry,
)

from . import io  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.core
~~~~~~~~~~~


"""

from .errors import (  # noqa: F401
    ResourceError,
    ResourceUnavailableError,
)

from .constants import Constants  # noqa: F401
from .constant import Constant, CONSTANTS  # noqa: F401

from .resource import Resource  # noqa: F401

from .resources import Resources  # noqa: F401

from .configs import (  # noqa: F401
    Directory,
    Directories,
    Configurations,
    Configurator,
    ConfigurationError,
    ConfigurationUnavailableError,
)

from .register import (  # noqa: F401
    Registry,
    Registrator,
    RegistratorContext,
    RegistratorAccess,
)

from .activator import (  # noqa: F401)
    Activator,
    activating,
)



# -*- coding: utf-8 -*-
"""
lories.connectors.sql.index
~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from enum import Enum
from typing import Collection, Optional

from sqlalchemy.types import DATE, DATETIME, INTEGER, TIME, TIMESTAMP, TypeEngine

import pytz as tz
from lories.connectors.sql.columns import Column, DatetimeColumn, UnixTimestampColumn


# noinspection PyShadowingBuiltins
class DatetimeIndexType(Enum):
    NONE = ()
    DATE_AND_TIME = (DATE, TIME)
    DATETIME = (DATETIME,)
    TIMESTAMP = (TIMESTAMP,)
    TIMESTAMP_UNIX = (INTEGER,)

    @classmethod
    def get(cls, type: str) -> DatetimeIndexType:
        for t in cls:
            if t.name == type:
                return t
        raise ValueError(type)

    def __contains__(self, type: TypeEngine):
        return any(type == t or isinstance(type, t) for t in self.value)

    def columns(self, name: Optional[str] = None) -> Collection[Column]:
        columns = []
        for column_type in self.value:
            column_name = name if name is not None else self.name.lower()
            if column_type == INTEGER:
                column = UnixTimestampColumn(column_name, column_type)
            else:
                column = DatetimeColumn(column_name, column_type, timezone=tz.UTC, nullable=False, primary_key=True)
            columns.append(column)
        return columns



# -*- coding: utf-8 -*-
"""
lories.components
~~~~~~~~~~~~~~~~~


"""

from . import context  # noqa: F401
from .context import (  # noqa: F401
    ComponentContext,
    register_component_type,
    registry,
)

from .errors import (  # noqa: F401
    ComponentError,
    ComponentUnavailableError,
)

from . import access  # noqa: F401
from .access import ComponentAccess  # noqa: F401

from . import component  # noqa: F401
from .component import Component  # noqa: F401

from . import weather  # noqa: F401
from .weather import (  # noqa: F401
    Weather,
    WeatherForecast,
)

from . import camera  # noqa: F401
from .camera import Camera  # noqa: F401

from . import tariff  # noqa: F401
from .tariff import Tariff  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.application.view.pages.layout.page
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from collections.abc import MutableSequence
from typing import Any, Collection, Optional

from dash_bootstrap_components import Container

from lories.application.view.pages.layout import PageCard


class PageLayout(MutableSequence[Any]):
    menu: Optional[Any]
    card: PageCard

    container: Container

    # noinspection PyPep8Naming
    def __init__(
        self,
        children: Collection[Any] = (),
        class_name: Optional[str] = None,
        className: Optional[str] = None,
        menu: Optional[Any] = None,
        **kwargs,
    ) -> None:
        if class_name is None or len(class_name) == 0:
            class_name = className
        if class_name is None or len(class_name) == 0:
            class_name = "page-container"
        else:
            class_name = f"{class_name} page-container"

        self.container = Container(
            children=list(children),
            class_name=class_name,
            **kwargs,
        )
        self.card = PageCard()
        self.menu = menu

    def __len__(self) -> int:
        return len(self.container)

    def __getitem__(self, index: int) -> Any:
        return self.container[index]

    def __setitem__(self, index: int, value: Any) -> None:
        self.container[index] = value

    def __delitem__(self, index: int) -> None:
        del self.container[index]

    def insert(self, index: int, value: Any) -> None:
        self.container.children.insert(index, value)

    def has_card_items(self) -> bool:
        return self.card.has_items()

    def has_menu_item(self) -> bool:
        return self.menu is not None



# -*- coding: utf-8 -*-
"""
lories.core.configs.configurator
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
from abc import ABCMeta
from collections import OrderedDict
from functools import wraps
from logging import Logger
from typing import Any, Dict, Optional

from lories._core import _Context, _Entity  # noqa
from lories._core._configurations import Configurations, _Configurations  # noqa
from lories._core._configurator import Configurator as ConfiguratorType  # noqa
from lories._core._configurator import _Configurator  # noqa
from lories.core.configs.errors import ConfigurationError
from lories.util import get_members


class ConfiguratorMeta(ABCMeta):
    def __call__(cls, *args, **kwargs):
        configurator = super().__call__(*args, **kwargs)
        cls._wrap_method(configurator, "duplicate")
        cls._wrap_method(configurator, "configure")
        cls._wrap_method(configurator, "update")

        return configurator

    # noinspection PyShadowingBuiltins
    @staticmethod
    def _wrap_method(object: Any, method: str) -> None:
        _wrap_method = getattr(object, f"_do_{method}")
        _run_method = getattr(object, method)
        setattr(object, f"_run_{method}", _run_method)
        setattr(object, method, _wrap_method)


class Configurator(_Configurator, metaclass=ConfiguratorMeta):
    __configs: _Configurations

    _configured: bool = False
    _logger: Logger

    def __init__(
        self,
        configs: Optional[Configurations] = None,
        logger: Optional[Logger] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        if logger is None:
            logger = self._build_logger()
        self._logger = logger
        self.__configs = self._assert_configs(configs)

    def __eq__(self, other: Any) -> bool:
        return self is other

    def __hash__(self) -> int:
        return hash(id(self))

    def __copy__(self) -> ConfiguratorType:
        return self.copy()

    def __replace__(self, **changes) -> ConfiguratorType:
        return self.duplicate(**changes)

    @classmethod
    def _assert_configs(cls, configs: Optional[Configurations]) -> Optional[Configurations]:
        if configs is None:
            return None
        if not isinstance(configs, _Configurations):
            raise ConfigurationError(f"Invalid '{cls.__name__}' configurations: {type(configs)}")
        return configs

    @classmethod
    def _build_logger(cls) -> Logger:
        return logging.getLogger(cls.__module__)

    def _get_vars(self) -> Dict[str, Any]:
        def _is_var(attr: str, var: Any) -> bool:
            return not (
                attr.startswith("_")
                or attr.isupper()
                or callable(var)
                or isinstance(var, _Context)
                or isinstance(var, _Configurations)
            )

        return get_members(self, filter=_is_var)

    # noinspection PyShadowingBuiltins
    def _convert_vars(self, convert: callable = str) -> Dict[str, str]:
        def _convert(var: Any) -> str:
            return str(var) if not isinstance(var, (_Context, _Configurator, _Entity)) else convert(var)

        vars = self._get_vars()
        values = OrderedDict([(k, _convert(v)) for k, v in vars.items()])
        if self.configs is not None:
            values["enabled"] = str(self.is_enabled())
            values["configured"] = str(self.is_configured())
            values["configs"] = convert(self.configs)
        return values

    # noinspection PyShadowingBuiltins
    def __repr__(self) -> str:
        vars = [f"{k}={v}" for k, v in self._convert_vars(lambda v: f"<{type(v).__name__}>").items()]
        return f"{type(self).__name__}({', '.join(vars)})"

    # noinspection PyShadowingBuiltins
    def __str__(self) -> str:
        vars = [f"{k} = {v}" for k, v in self._convert_vars(repr).items()]
        return f"{type(self).__name__}:\n\t" + "\n\t".join(vars)

    def is_enabled(self) -> bool:
        return self.__configs is not None and self.__configs.enabled

    def is_configured(self) -> bool:
        return self._configured

    @property
    def configs(self) -> Optional[Configurations]:
        return self.__configs

    # noinspection PyUnresolvedReferences
    @wraps(_Configurator.configure, updated=())
    def _do_configure(self, configs: Configurations, *args, **kwargs) -> None:
        if configs is None:
            raise ConfigurationError(f"Invalid NoneType configuration for {type(self).__name__}: {self.name}")
        if not configs.enabled:
            raise ConfigurationError(f"Trying to configure disabled {type(self).__name__}: {configs.name}")
        if self.is_configured():
            self._logger.warning(f"{type(self).__name__} '{configs.path}' already configured")
            return

        self._assert_configs(configs)
        self._at_configure(configs)
        self._run_configure(configs, *args, **kwargs)
        self._on_configure(configs)
        self.__configs = configs
        self._configured = True

    def _at_configure(self, configs: Configurations) -> None:
        pass

    def _on_configure(self, configs: Configurations) -> None:
        pass

    # noinspection PyUnresolvedReferences
    def update(self, configs: Configurations) -> None:
        self._run_configure(configs)

    # noinspection PyUnresolvedReferences
    @wraps(_Configurator.update, updated=())
    def _do_update(self, configs: Configurations, *args, **kwargs) -> None:
        if configs is None:
            raise ConfigurationError(f"Invalid NoneType configuration for {type(self).__name__}: {self.name}")
        if not configs.enabled:
            raise ConfigurationError(f"Trying to update disabled {type(self).__name__}: {configs.name}")
        if not self.is_configured():
            self._logger.warning(f"Trying to update unconfigured {type(self).__name__}: '{configs.path}'")
            return

        self._assert_configs(configs)
        self._at_update(configs)
        self._run_update(configs, *args, **kwargs)
        self._on_update(configs)
        self.__configs = configs

    def _at_update(self, configs: Configurations) -> None:
        pass

    def _on_update(self, configs: Configurations) -> None:
        pass

    # noinspection PyUnresolvedReferences, PyTypeChecker
    def copy(self) -> ConfiguratorType:
        try:
            copier = super().copy
        except AttributeError:
            copier = self.duplicate
        return copier()

    # noinspection PyUnresolvedReferences
    def duplicate(self, configs: Optional[Configurations] = None, **changes) -> ConfiguratorType:
        if configs is None:
            configs = self.__configs.copy()
        try:
            duplicator = super().duplicate
        except AttributeError:
            duplicator = type(self)
        return duplicator(configs=configs, **changes)

    # noinspection PyUnresolvedReferences
    @wraps(_Configurator.duplicate, updated=())
    def _do_duplicate(self, configs: Configurations, **changes) -> ConfiguratorType:
        if configs is None:
            configs = self.__configs.copy()
        self._at_duplicate(configs=configs, **changes)

        duplicate = self._run_duplicate(configs=configs, **changes)
        if configs.enabled:
            duplicate.configure(configs)
        self._on_duplicate(duplicate)
        return duplicate

    def _at_duplicate(self, **changes) -> None:
        pass

    def _on_duplicate(self, configurator: Configurator) -> None:
        pass



# -*- coding: utf-8 -*-
"""
lories.connectors.sql.schema
~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Collection, Dict, Iterable

from sqlalchemy import Connection, Dialect, Engine, ForeignKey, MetaData, inspect

import pytz as tz
from lories.connectors.sql.columns import (
    Column,
    DatetimeColumn,
    SurrogateKeyColumn,
    parse_type,
)
from lories.connectors.sql.columns.datetime import is_datetime
from lories.connectors.sql.index import DatetimeIndexType
from lories.connectors.sql.table import Table
from lories.core import ConfigurationError, Configurator, ResourceError
from lories.typing import Configurations, Resource, Resources
from lories.util import to_bool, to_timezone


class Schema(Configurator, MetaData):
    dialect: Dialect

    _created: bool = False

    def __init__(self, dialect: Dialect, **kwargs) -> None:
        super().__init__(**kwargs)
        self._created = False
        self.dialect = dialect

    def __repr__(self) -> str:
        # Do not use __repr__ of Configurator class, to avoid infinite recursion
        return super(MetaData, self).__repr__()

    def __str__(self) -> str:
        # Do not use __str__ of Configurator class, to avoid infinite recursion
        return super(MetaData, self).__str__()

    def connect(self, bind: Engine | Connection, resources: Resources) -> Dict[str, Table]:
        tables = self._connect_tables(resources)
        if not self._created:
            self.create_all(bind=bind, checkfirst=True)
            self._created = True
        self._validate(bind=bind, tables=tables.values())
        return tables

    def _connect_tables(self, resources: Resources) -> Dict[str, Table]:
        tables = {}

        defaults = self.configs.get_members(["index", "columns"], ensure_exists=True)
        for schema, schema_resources in resources.groupby("schema"):
            for name, table_resources in schema_resources.groupby(lambda c: c.get("table", default=c.group)):
                if name is None:
                    raise ConfigurationError(
                        "Missing 'table' configuration for resources: " + ", ".join([r.id for r in table_resources])
                    )
                if name in self.tables:
                    table = self.tables[name]
                    tables[table.key] = table
                    continue

                configs = self.configs.get_member(name, defaults=defaults)
                columns_configs = configs.get_member("columns")
                column_configs = [columns_configs[s] for s in columns_configs.members]

                def _filter_primary(primary: bool) -> Collection[Resource | Configurations]:
                    _filter_columns = [*column_configs, *table_resources]
                    return [r for r in _filter_columns if r.get("primary", default=False) == primary]

                columns = []
                columns.extend(Schema._create_primary_key(configs.get_member("index"), *_filter_primary(True)))
                columns.extend(Schema._create_columns(*_filter_primary(False)))

                def _filter_duplicates() -> Collection[Column]:
                    unique = set()
                    return [c for c in columns if (c.name in unique or unique.add(c.name))]

                # Remove duplicate columns that will be created for tables with surrogate keys,
                # which will have duplicate resource configurations.
                for duplicate in _filter_duplicates():
                    if not any(self._validate_column(column, duplicate) for column in columns):
                        raise ConfigurationError(f"Duplicate column for table '{name}': {duplicate}")
                    columns.remove(duplicate)

                table = Table(name, self, *columns, schema=schema, quote=True, quote_schema=True)
                tables[table.key] = table
        return tables

    # noinspection PyShadowingBuiltins
    @staticmethod
    def _create_primary_key(configs: Configurations, *resources: Resource | Configurations) -> Iterable[Column]:
        columns = []
        type = configs.get("type", default="default")
        if type.lower() in ["default", "none"]:
            type = DatetimeIndexType.TIMESTAMP
            columns = type.columns(configs.get("column", default=None))
        elif type.lower() != "custom":
            type = DatetimeIndexType.get(type.upper())
            columns = type.columns(configs.get("column", default=None))

        columns.extend(Schema._create_columns(*resources))
        return columns

    @staticmethod
    def _create_columns(*resources: Resource | Configurations) -> Iterable[Column]:
        return [Schema._create_column(resource) for resource in resources]

    # noinspection PyShadowingBuiltins
    @staticmethod
    def _create_column(resource: Resource | Configurations) -> Column:
        name = resource.get("column", default=resource.key)
        type = parse_type(resource["type"], resource.get("length", None))
        primary = to_bool(resource.get("primary", default=False))
        configs = {
            "primary_key": primary,
            "default": resource.get("default", default=None),
            "nullable": resource.get("nullable", default=True if not primary else False),
        }
        if primary:
            relations = []
            foreign = resource.get("foreign", default=None)
            if foreign is not None and len(foreign) > 0:
                relations.append(ForeignKey(foreign, ondelete="CASCADE"))

            attribute = resource.get("attribute", default=None)
            if attribute is not None and len(attribute) > 0:
                return SurrogateKeyColumn(name, type, attribute, *relations, **configs)

        if is_datetime(type):
            configs["timezone"] = to_timezone(resource.get("timezone", default=tz.UTC)) if not primary else tz.UTC
            return DatetimeColumn(name, type, **configs)

        return Column(name, type, **configs)

    def _validate(self, bind: Engine | Connection, tables: Collection[Table]) -> None:
        self.reflect(bind)

        inspector = inspect(bind)
        for table in tables:
            columns = inspector.get_columns(table.name, table.schema)
            column_names = [c["name"] for c in columns]
            for column in table.columns.values():
                if column.name in column_names:
                    column_schema = next(c for c in columns if c["name"] == column.name)  # noqa F841
                    # TODO: Implement column validation
                else:
                    if self.configs.get_bool("create", default=True):
                        raise ResourceError(f"Not yet implemented to create missing column: {column.name}")
                    else:
                        raise ResourceError(f"Unable to find configured column: {column.name}")

    @staticmethod
    def _validate_column(column: Column, other: Column) -> bool:
        return (
            column.name == other.name
            and column.type == other.type
            and column.nullable == column.nullable
            and column.primary_key == other.primary_key
            and column.server_default == column.server_default
            and column.server_onupdate == column.server_onupdate
        )



# -*- coding: utf-8 -*-
"""
lories.data.channels
~~~~~~~~~~~~~~~~~~~~


"""

from ..._core import ChannelState  # noqa: F401

from .connector import ChannelConnector  # noqa: F401
from .converter import ChannelConverter  # noqa: F401

from .channels import Channels  # noqa: F401

from .channel import Channel  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories._core._resource
~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import abstractmethod
from typing import Any, Dict, Optional, Sequence, Type, TypeVar

from lories._core._entity import _Entity


class _Resource(_Entity):
    @abstractmethod
    def full_name(self, unit: bool = False) -> str: ...

    @property
    @abstractmethod
    def group(self) -> str: ...

    @property
    @abstractmethod
    def unit(self) -> Optional[str]: ...

    @property
    @abstractmethod
    def type(self) -> Type: ...

    @abstractmethod
    def get(self, attr: str, default: Optional[Any] = None) -> Any: ...

    @abstractmethod
    def to_list(self) -> Sequence[Resource]: ...

    @abstractmethod
    def to_configs(self) -> Dict[str, Any]: ...


Resource = TypeVar("Resource", bound=_Resource)



# -*- coding: utf-8 -*-
"""
lories.application.view
~~~~~~~~~~~~~~~~~~~~~~~


"""

from . import login  # noqa: F401
from .login import (  # noqa: F401
    Authentication,
    AuthenticationException,
    LoginPage,
)

from . import pages  # noqa: F401
from .pages import (  # noqa: F401
    Page,
    PageGroup,
    PageHeader,
    PageFooter,
    ComponentPage,
    ComponentGroup,
    View,
)

from .interface import ViewInterface  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.core.configs.configurations
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import os
import re
import shutil
import tempfile
from collections import OrderedDict
from copy import deepcopy
from pathlib import Path
from typing import Any, Collection, Iterable, Iterator, List, Mapping, Optional

import pandas as pd
from lories._core import _Configurations  # noqa
from lories._core.typing import Timestamp  # noqa
from lories.core.configs.directories import Directories, Directory
from lories.core.configs.errors import ConfigurationError, ConfigurationUnavailableError
from lories.util import is_bool, to_bool, to_date, to_float, to_int, update_recursive


class Configurations(_Configurations):
    @classmethod
    def load(
        cls,
        conf_file: str,
        conf_dir: str = None,
        data_dir: str = None,
        tmp_dir: str = None,
        log_dir: str = None,
        lib_dir: str = None,
        flat: bool = False,
        require: bool = True,
        **defaults,
    ) -> Configurations:
        if not conf_dir and flat:
            conf_dir = ""

        conf_dirs = Directories(lib_dir, log_dir, tmp_dir, data_dir, conf_dir)
        conf_path = Path(conf_dirs.conf, conf_file)

        if conf_dirs.conf.is_dir():
            if not conf_path.is_file():
                config_default = str(conf_path).replace(".conf", ".default.conf")
                if os.path.isfile(config_default):
                    shutil.copy(config_default, conf_path)
        elif require:
            raise ConfigurationUnavailableError(f"Invalid configuration directory: {conf_dirs.conf}")

        configs = cls(conf_file, conf_dirs, defaults)
        configs._load(require)
        return configs

    # noinspection PyProtectedMember
    def _load(self, require: bool = True) -> None:
        if self.__path.exists() and self.__path.is_file():
            try:
                # TODO: Implement other configuration parsers
                self._load_toml(str(self.__path))
            except Exception as e:
                raise ConfigurationUnavailableError(f"Error loading configuration file '{self.__path}': {str(e)}")

        elif require:
            raise ConfigurationUnavailableError(f"Invalid configuration file '{self.__path}'")

        if Directories.TYPE in self.__configs:
            self.__dirs.update(self.__configs[Directories.TYPE])

    def _load_toml(self, config_path: str) -> None:
        from .toml import load_toml

        self.update(load_toml(config_path))

    def __init__(
        self,
        name: str,
        dirs: Directories,
        defaults: Optional[Mapping[str, Any]] = None,
        **kwargs,
    ) -> None:
        super().__init__()
        self.__configs = OrderedDict()
        self.__dirs = dirs
        self.__path = Path(dirs.conf, name)

        if defaults is not None:
            self.update(defaults)
        self.update(kwargs)

    def __repr__(self) -> str:
        return f"{Configurations.__name__}({self.__path})"

    def __str__(self) -> str:
        # noinspection PyShadowingNames
        def parse_configs(header: str, configs: Configurations) -> str:
            configs = OrderedDict(configs)
            for k in [k for k, c in configs.items() if isinstance(c, Mapping)]:
                configs.move_to_end(k)

            string = f"[{header}]\n"
            for k, v in configs.items():
                if isinstance(v, Configurations):
                    string += "\n" + parse_configs(f"{header}.{k}", v)
                else:
                    string += f"{k} = {v}\n"
            return string

        return parse_configs(self.name.replace(".conf", ""), self)

    def __delitem__(self, key: str) -> None:
        del self.__configs[key]

    def remove(self, *keys: str) -> None:
        for key in keys:
            del self.__configs[key]

    def pop(self, key: str, default: Any = None) -> Any:
        value = self._get(key, default)
        if key in self.__configs:
            del self.__configs[key]
        return value

    def __setitem__(self, key: str, value: Any) -> None:
        self.set(key, value)

    # noinspection PyTypeChecker
    def set(self, key: str, value: Any, replace: bool = True) -> None:
        if isinstance(value, Mapping):
            if key not in self.__configs.keys():
                value = self._create_member(key, value)
            else:
                _key = self.__configs[key]
                if isinstance(_key, Mapping) and not replace:
                    value = update_recursive(_key, value, replace=False)
            self.__configs[key] = value
        elif key not in self.__configs.keys() or replace:
            self.__configs[key] = value

    def __getitem__(self, key: str) -> Any:
        return self.__configs[key]

    def _get(self, key: str, default: Any = None) -> Any:
        return self.__configs.get(key, default)

    def get(self, key: str | Iterable[str], default: Any = None) -> Any:
        if not isinstance(key, Iterable) or isinstance(key, str):
            return self._get(key, default)
        return {
            k: self._get(k, default=default[k] if default is not None and isinstance(default, Mapping) else None)
            for k in key
            if k in self
        }

    def get_bool(self, key: str, default: bool = None) -> bool:
        return to_bool(self._get(key, default))

    def get_int(self, key: str, default: int = None) -> int:
        return to_int(self._get(key, default))

    def get_float(self, key: str, default: float = None) -> float:
        return to_float(self._get(key, default))

    def get_date(self, key: str, default: Timestamp = None, **kwargs) -> pd.Timestamp:
        return to_date(self._get(key, default), **kwargs)

    def __contains__(self, key: str) -> bool:
        return key in self.__configs

    def __iter__(self) -> Iterator[str]:
        return iter(self.__configs)

    def __len__(self) -> int:
        return len(self.__configs)

    def move_to_top(self, key: str) -> None:
        self.__configs.move_to_end(key, False)

    def move_to_bottom(self, key: str) -> None:
        self.__configs.move_to_end(key, True)

    def write(self) -> None:
        configs = {k: v for k, v in self.__configs.items() if k not in self.members}

        if not self.__dirs.conf.exists():
            self.__dirs.conf.mkdir(parents=True, exist_ok=True)

        file_desc, file_path = tempfile.mkstemp(prefix=self.name, dir=self.dirs.conf)
        with os.fdopen(file_desc, "w") as file:
            lines = self.__read_lines()
            lines_member = len(lines) - 1
            for line_index, line in enumerate(lines):
                if "=" in line:
                    line = line.rstrip()
                    key, value, *_ = line.split("=")
                    key = key.lstrip().lstrip("#").lstrip(";").strip()
                    value = value.strip().strip('"')
                    if key in configs:
                        config_value = str(configs.pop(key))
                        if config_value.lower() != value.lower() or line.lstrip().startswith(("#", ";")):
                            lines[line_index] = self.__parse_line(key, config_value)

                if re.match(r"(#.*|;.*|)\[.*?]", line):
                    lines_member = line_index - 1
                    break
            while lines_member > 0 and lines[lines_member - 1].strip() == "":
                lines_member -= 1

            if len(configs) > 0:
                if len(lines) > 0:
                    lines.insert(lines_member, "\n")
                    lines_member += 1
                for key, value in configs.items():
                    lines.insert(lines_member, self.__parse_line(key, value))
                    lines_member += 1

            file.writelines(lines)

        # Copy the file permissions from the configuration file to the temporary file and remove it
        if self.__path.exists():
            shutil.copymode(self.__path, file_path)
            os.remove(self.__path)

        shutil.move(file_path, self.__path)

    def __read_lines(self) -> List[str]:
        if not self.__path.exists():
            return []
        with open(self.__path, "r") as file:
            return file.readlines()

    # noinspection PyMethodMayBeStatic
    def __parse_line(self, key, value: Any) -> str:
        if is_bool(value):
            value = str(value).lower()
        elif isinstance(value, str):
            if "\\" in value:
                value = value.translate(str.maketrans({"\\": r"\\"}))
            value = f'"{value}"'
        return f"{key} = {value}\n"

    def copy(self, dirs: Optional[Directories] = None) -> Configurations:
        if dirs is None:
            dirs = deepcopy(self.dirs)
        elif dirs.conf != self.dirs.conf:
            self.__copy_path(self.__path.parents[0], dirs.conf, self.name)
            self.__copy_path(self.__path.parents[0], dirs.conf, self.name.replace(".conf", ".d"))
            for member in self.members:
                member_dir = dirs.conf.joinpath(self.name.replace(".conf", ".d"))
                self.__copy_path(self.__path.parents[0], member_dir, f"{member}.conf")

        return Configurations(self.name, dirs, deepcopy(self.__configs))

    @staticmethod
    def __copy_path(source: Path, destination: Path, name: str) -> None:
        source = source.joinpath(name)
        destination = destination.joinpath(name)
        if not source.exists():
            return

        destination.parents[0].mkdir(parents=True, exist_ok=True)
        if source.is_dir():
            shutil.copytree(source, destination, ignore=_include(r".*\.conf"), dirs_exist_ok=True)
        elif not destination.exists():
            shutil.copy2(source, destination)

    @property
    def key(self) -> str:
        return str(self.__path.name.removesuffix(".conf"))

    @property
    def name(self) -> str:
        return str(self.__path.name)

    @property
    def path(self) -> str:
        return str(self.__path)

    @property
    def dirs(self) -> Directories:
        return self.__dirs

    @property
    def enabled(self) -> bool:
        return to_bool(self._get("enabled", default=True)) and not to_bool(self._get("disabled", default=False))

    @enabled.setter
    def enabled(self, enabled: bool) -> None:
        self.set("enabled", enabled)

    @property
    def members(self) -> List[str]:
        return [k for k, v in self.items() if isinstance(v, Configurations)]

    @property
    def _members_dir(self) -> Directory:
        return self.__dirs.conf.joinpath(self.__path.name.replace(".conf", ".d"))

    def has_member(self, key: str, includes: bool = False) -> bool:
        if key in self.members:
            return True
        if includes and self._members_dir.joinpath(f"{key}.conf").exists():
            return True
        return False

    def get_members(
        self,
        keys: Collection[str],
        ensure_exists: bool = False,
    ) -> Configurations:
        member = {
            s: self.get_member(s, defaults={}, ensure_exists=ensure_exists)
            for s in keys
            if s in self.members or ensure_exists
        }
        member_dirs = self.__dirs.copy()
        member_dirs.conf = self._members_dir
        return Configurations(self.name, member_dirs, member)

    def get_member(
        self,
        key: str,
        defaults: Optional[Mapping[str, Any]] = None,
        ensure_exists: bool = False,
    ) -> Configurations:
        if not self.has_member(key) and ensure_exists:
            if defaults is None:
                defaults = {}
            self._add_member(key, defaults)
            return self[key]

        elif self.has_member(key):
            configs = self[key]
            if defaults is not None:
                configs.update(defaults, replace=False)
            return configs

        elif defaults is not None:
            return self._create_member(key, defaults)
        else:
            raise ConfigurationUnavailableError(f"Unknown configuration type '{key}'")

    def _add_member(self, key, configs: Mapping[str, Any]) -> None:
        if self.has_member(key):
            raise ConfigurationUnavailableError(f"Unable to add existing configuration type '{key}'")
        self[key] = self._create_member(key, configs)

    def _create_member(self, key, configs: Mapping[str, Any]) -> Configurations:
        if not isinstance(configs, Mapping):
            raise ConfigurationError(f"Invalid configuration type '{key}': {type(configs)}")
        member_name = f"{key}.conf"
        member_dirs = self.__dirs.copy()
        member_dirs.conf = self._members_dir
        member_configs = Configurations(member_name, member_dirs, configs)
        member_configs._load(require=False)
        return member_configs

    def pop_member(
        self,
        key: str,
        defaults: Optional[Mapping[str, Any]] = None,
    ) -> Configurations:
        member_configs = self.get_member(key, defaults=defaults)
        if key in self.__configs:
            del self.__configs[key]
        return member_configs

    # noinspection PyTypeChecker
    def update(self, update: Mapping[str, Any], replace: bool = True) -> None:
        update_recursive(self, update, replace=replace)


def _include(pattern):
    def _ignore(path, names):
        return set(n for n in names if not re.match(pattern, n) and not os.path.isdir(os.path.join(path, n)))

    return _ignore



# -*- coding: utf-8 -*-
"""
lories.connectors.influx
~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
from typing import Any, Dict, Optional, Tuple

from influxdb_client import BucketRetentionRules, InfluxDBClient
from influxdb_client.client.exceptions import InfluxDBError
from influxdb_client.client.write_api import SYNCHRONOUS
from influxdb_client.rest import ApiException
from urllib3.exceptions import HTTPError, NewConnectionError

import pandas as pd
from lories.connectors import ConnectionError, Database, DatabaseException, register_connector_type
from lories.core.configs import ConfigurationError
from lories.data.util import hash_value
from lories.typing import Configurations, Resource, Resources, Timestamp

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


@register_connector_type("influx", "influxdb")
class InfluxDatabase(Database):
    host: str
    port: int

    org: str
    bucket: str
    token: str

    ssl: bool
    ssl_verify: bool
    timeout: int

    _client: Optional[InfluxDBClient] = None

    # noinspection PyShadowingBuiltins
    def _get_vars(self) -> Dict[str, Any]:
        vars = super()._get_vars()
        if self.is_configured():
            vars["host"] = self.host
            vars["port"] = self.port
            vars["org"] = self.org
            vars["bucket"] = self.bucket
            vars["token"] = f"...{self.token[-6:]}"
            vars["ssl"] = self.ssl_verify
            vars["ssl_verify"] = self.ssl_verify
            vars["timeout"] = self.timeout
        return vars

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)

        self.host = configs.get("host", default="localhost")
        self.port = configs.get_int("port", default=8086)

        self.org = configs.get("org")
        if self.org is None:
            raise ConfigurationError("Missing 'org' for InfluxDB connector")

        self.bucket = configs.get("bucket")
        if self.bucket is None:
            raise ConfigurationError("Missing 'bucket' for InfluxDB connector")

        self.token = configs.get("token")
        if self.token is None:
            raise ConfigurationError("Missing 'token' for InfluxDB connector")

        # In seconds
        self.timeout = int(configs.get_float("timeout", default=10) * 1000)

        # TODO: Determine SSL usage from certificate availability
        self.ssl = configs.get_bool("ssl", default=False)
        self.ssl_verify = configs.get_bool("ssl_verify", default=True)

    def connect(self, resources: Resources) -> None:
        self._logger.debug(f"Connecting to InfluxDB ({self.host}:{self.port}) at {self.bucket}")

        ssl = {
            "verify_ssl": self.ssl_verify,
        }
        url = f"{'https' if self.ssl else 'http'}://{self.host}:{self.port}"

        self._client = InfluxDBClient(
            url=url,
            org=self.org,
            token=self.token,
            timeout=self.timeout,
            debug=self._logger.getEffectiveLevel() <= logging.DEBUG,
            **ssl,
        )

        # Check if bucket exists and create
        buckets_api = self._client.buckets_api()
        try:
            existing_buckets = [b.name for b in buckets_api.find_buckets().buckets]
            if self.bucket not in existing_buckets:
                self._logger.info(f"InfluxDB Bucket doesn't exist. Creating Bucket {self.bucket}")
                # TODO: Configure retention from channel attributes
                retention = BucketRetentionRules(every_seconds=0)  # 0 means infinite retention
                buckets_api.create_bucket(bucket_name=self.bucket, retention_rules=retention)
        except (ApiException, InfluxDBError, HTTPError) as e:
            self._raise(e)

    def disconnect(self) -> None:
        # TODO: Check if "and ping" is necessary
        if self._client is not None and self._client.ping():
            self._client.close()
            self._client = None

    def hash(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
        method: Literal["MD5", "SHA1", "SHA256", "SHA512"] = "MD5",
        encoding: str = "UTF-8",
    ) -> Optional[str]:
        if method.lower() not in ["md5", "sha1", "sha256"]:
            raise ValueError(f"Invalid checksum method '{method}'")
        hashes = []

        imports = """
        import "types" // = [string, bytes, int, uint, !(float), bool, time, duration, regexp]
        import "array"
        import "strings"
        import "math"
        import "contrib/qxip/hash"
        """

        # functions = f"""
        # roundDigits = (n, dig) =>
        #     string(v: math.round(x: float(v: n)
        #     * float(v: math.pow10(n: dig)))
        #     / float(v: math.pow10(n: dig)))
        #
        # formatValue = (n, dig) =>
        #     if exists n then
        #         if types.isType(v: n, type: "float") then
        #             roundDigits(n: n, dig: dig) + ","
        #             // if strings.containsStr(substr: ".", v: roundDigits(n: n, dig: dig)) then
        #             //     roundDigits(n: n, dig: dig) + ","
        #             // else
        #             //     roundDigits(n: n, dig: dig) + ".000" + ","
        #         else
        #             string(v: n) + ","
        #     else
        #         ""
        # """
        functions = """
        formatValue = (n) =>
            if exists n then
                string(v: n) + ","
            else
                ""
        """

        query_api = self._client.query_api()
        for measurement, measurement_resources in resources.groupby(lambda r: r.get("measurement", default=r.group)):
            for tag, tagged_resources in measurement_resources.groupby("tag"):
                # decimals = 3
                # concat = "            + ".join(
                #     f"formatValue(n: r.{_get_field(r)}, dig: {r.get('decimals', default=decimals)})"
                #     for r in tagged_resources
                # )
                concat = " + ".join(f"formatValue(n: r.{_get_field(r)})" for r in tagged_resources)

                query = f"""
                {imports}

                {functions}

                {self._build_query(tagged_resources, measurement, tag, *_to_isoformat(start, end))}
                    |> pivot(rowKey:["_time"], columnKey: ["_field"], valueColumn: "_value")
                    |> map(fn: (r) => ({{
                        r with
                        row_strings: string(v: int(v: r._time) / 1000000000) + ","
                            + {concat}
                        }}))
                    |> reduce(identity: {{column_string: ""}},
                        fn: (r, accumulator) => ({{
                            column_string: accumulator.column_string + r.row_strings
                        }}))
                    |> map(fn: (r) => ({{
                        r with
                        hash: hash.{method.lower()}(v: r.column_string)
                        // hash: strings.substring(v: r.column_string, start: 0, end: 300)
                        }}))
                    |> keep(columns: ["hash"])
                """
                try:
                    result = query_api.query_data_frame(query)
                    if result.empty:
                        continue
                    hashes.append(result["hash"].iloc[0])

                except (ApiException, InfluxDBError, HTTPError) as e:
                    self._raise(e)

        if len(hashes) == 0:
            return None
        elif len(hashes) == 1:
            return hashes[0]
        else:
            return hash_value(",".join(hashes), method, encoding)

    def exists(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> bool:
        query_api = self._client.query_api()
        for measurement, measurement_resources in resources.groupby(lambda r: r.get("measurement", default=r.group)):
            for tag, tagged_resources in measurement_resources.groupby("tag"):
                fields = [_get_field(r) for r in tagged_resources]
                query = f"""
                {self._build_query(tagged_resources, measurement, tag, *_to_isoformat(start, end))}
                    |> keep(columns: ["_field"])
                    |> distinct(column: "_field")
                """
                try:
                    data = query_api.query_data_frame(query, data_frame_index=["_time"])
                    if sorted(data["_fields"]) != sorted(fields):
                        return False

                except (ApiException, InfluxDBError, HTTPError) as e:
                    self._raise(e)
        return True

    # noinspection PyUnresolvedReferences, PyTypeChecker
    def read(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> pd.DataFrame:
        return self._read(resources, start, end)

    # noinspection PyUnresolvedReferences, PyTypeChecker
    def read_first(self, resources: Resources) -> pd.DataFrame:
        first = self._read_boundaries(resources, "first")
        return first

    # noinspection PyUnresolvedReferences, PyTypeChecker
    def read_last(self, resources: Resources) -> pd.DataFrame:
        last = self._read_boundaries(resources, "last")
        return last

    # noinspection PyUnresolvedReferences, PyTypeChecker
    def _read_boundaries(self, resources: Resources, mode: Literal["first", "last"]) -> pd.DataFrame:
        if mode not in ["first", "last"]:
            raise ValueError(f"Invalid mode '{mode}'")
        results = []

        query_api = self._client.query_api()
        for measurement, measurement_resources in resources.groupby(lambda r: r.get("measurement", default=r.group)):
            for tag, tagged_resources in measurement_resources.groupby("tag"):
                query = f"""
                {self._build_query(tagged_resources, measurement, tag, *_to_isoformat(None, None))}
                    |> group(columns: ["_field"])
                    |> {mode}(column: "_time")
                    |> pivot(rowKey:["_time"], columnKey: ["_field"], valueColumn: "_value")
                """
                try:
                    data = query_api.query_data_frame(query, data_frame_index=["_time"])
                    if not data.empty:
                        results.append(
                            data.rename(
                                columns={_get_field(r): r.id for r in tagged_resources},
                            )
                        )

                except (ApiException, InfluxDBError, HTTPError) as e:
                    self._raise(e)

        if len(results) == 0:
            return pd.DataFrame(columns=[r.id for r in resources])

        results = pd.concat(results, axis="columns")
        results.sort_index(inplace=True)
        results = results.loc[:, [r.id for r in resources if r.id in results.columns]]
        return results

    # noinspection PyUnresolvedReferences, PyTypeChecker
    def _read(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> pd.DataFrame:
        results = []

        query_api = self._client.query_api()
        for measurement, measurement_resources in resources.groupby(lambda r: r.get("measurement", default=r.group)):
            for tag, tagged_resources in measurement_resources.groupby("tag"):
                query = f"""
                    {self._build_query(tagged_resources, measurement, tag, *_to_isoformat(start, end))}
                    |> pivot(rowKey:["_time"], columnKey: ["_field"], valueColumn: "_value")
                """
                try:
                    data = query_api.query_data_frame(query, data_frame_index=["_time"])

                    if not data.empty:
                        results.append(
                            data.rename(
                                columns={_get_field(r): r.id for r in tagged_resources},
                            )
                        )

                except (ApiException, InfluxDBError, HTTPError) as e:
                    self._raise(e)

        if len(results) == 0:
            return pd.DataFrame(columns=[r.id for r in resources])

        results = pd.concat(results, axis="columns")
        results.sort_index(inplace=True)
        results = results.loc[:, [r.id for r in resources if r.id in results.columns]]
        return results

    # noinspection PyTypeChecker
    def write(self, data: pd.DataFrame) -> None:
        write_api = self._client.write_api(write_options=SYNCHRONOUS)
        for measurement, group_resources in self.resources.groupby(lambda r: r.get("measurement", default=r.group)):
            if measurement is None:
                measurement = "None"
            for tag, tagged_resources in group_resources.groupby("tag"):
                tagged_data = data.loc[:, [r.id for r in tagged_resources if r.id in data.columns]]
                tagged_data = tagged_data.dropna(axis="index", how="all").dropna(axis="columns", how="all")
                if tagged_data.empty:
                    continue

                if tag is not None:
                    tagged_data["tag"] = tag

                try:
                    tagged_data = tagged_data.rename(columns={r.id: _get_field(r) for r in tagged_resources})
                    write_api.write(
                        bucket=self.bucket,
                        record=tagged_data,
                        data_frame_measurement_name=measurement,
                        data_frame_tag_columns=["tag"] if tag is not None else None,
                    )
                    write_api.flush()

                except (ApiException, InfluxDBError, HTTPError) as e:
                    self._raise(e)

    def delete(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> None:
        start, end = _to_isoformat(start, end)
        delete_api = self._client.delete_api()
        for measurement, group_resources in resources.groupby(lambda r: r.get("measurement", default=r.group)):
            for tag, tagged_resources in group_resources.groupby("tag"):
                # Escaped because "tag" is a keyword
                predicate = f'_measurement = "{measurement}"'
                predicate += f' AND "tag" = "{tag}"' if tag is not None else ""

                try:
                    delete_api.delete(
                        start=start,
                        stop=end,
                        predicate=predicate,
                        bucket=self.bucket,
                        org=self.org,
                    )
                except (ApiException, InfluxDBError, HTTPError) as e:
                    self._raise(e)

    def _build_query(
        self,
        resources: Resources,
        measurement: str,
        tag: Optional[str],
        start: str = "0",
        end: str = "now()",
    ) -> str:
        if measurement is None:
            raise DatabaseException(self, f"Measurement is None for following resources: {resources}")
        query = f"""
            from(bucket: "{self.bucket}")
                |> range(start: {start}, stop: {end})
                |> filter(fn: (r) => r["_measurement"] == "{measurement}")
                |> filter(fn: (r) => {" or ".join([f'r["_field"] == "{_get_field(r)}"' for r in resources])})
            """
        if tag is not None:
            query += f"""
                |> filter(fn: (r) => r["tag"] == "{tag}")
            """
        return query

    # noinspection PyProtectedMember
    def is_connected(self) -> bool:
        if self._client is None:
            return False
        try:
            return self._client.ping()
        except (ApiException, InfluxDBError, HTTPError) as e:
            self._raise(e)
        return False

    def _raise(self, e: Exception):
        if isinstance(e, ApiException):
            raise DatabaseException(self, str(e))
        elif isinstance(e, InfluxDBError):
            raise DatabaseException(self, str(e))
        elif isinstance(e, NewConnectionError):
            raise DatabaseException(self, str(e))
        else:
            raise ConnectionError(self, str(e))


# noinspection SpellCheckingInspection
def _to_isoformat(
    start: Optional[Timestamp] = None,
    end: Optional[Timestamp] = None,
) -> Tuple[str, str]:
    if start is None:
        start = "0"
    else:
        start = start.isoformat(sep="T", timespec="seconds")
    if end is None:
        end = "now()"
    else:
        end += pd.Timedelta(seconds=1)
        end = end.isoformat(sep="T", timespec="seconds")
    return start, end


def _get_field(resource: Resource) -> str:
    return resource.get("field", default=resource.get("column", default=resource.key))



# -*- coding: utf-8 -*-
"""
lories.data.retention
~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
from typing import Any, Dict, Iterable, Iterator, MutableSequence, Optional

import tzlocal

import pandas as pd
import pytz as tz
from lories import ConfigurationError, Configurations, Resource, ResourceError, Resources
from lories.data.util import hash_data, resample
from lories.util import floor_date, parse_freq, slice_range, to_bool, to_timedelta, to_timezone

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal, LiteralString

except ImportError:
    from typing_extensions import Literal, LiteralString


class Retention:
    TYPE: str = "retention"

    _enabled: bool = False

    keep: str
    freq: str
    method: Optional[LiteralString]
    resample: Optional[str]
    timezone: tz.BaseTzInfo

    # noinspection PyShadowingNames
    @classmethod
    def build(cls, configs: Configurations, resource: Resource) -> Retentions:
        resource_configs = resource.get(cls.TYPE, default={})
        if isinstance(resource_configs, str):
            resource_configs = {"aggregate": resource_configs}
        elif isinstance(resource_configs, Dict):
            resource_configs["aggregate"] = resource.get("aggregate", default=None)
        else:
            raise ConfigurationError("Invalid retention method: " + str(resource_configs))

        configs = configs.copy()
        configs.update(resource_configs)
        defaults = {k: configs.pop(k) for k in list(configs.keys()) if k not in configs.members}

        retentions = Retentions()
        for retain in [r for r in configs.members if r not in ["databases"]]:
            retention = configs.get_member(retain, defaults=defaults)
            retain = retention.pop("keep", default=retain)
            retention = cls(retain, **retention)
            if retention.enabled:
                retentions.append(retention)
        return retentions

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        keep: str,
        freq: str = "D",
        resample: Optional[str] = None,
        aggregate: Optional[Literal["sum", "mean", "min", "max", "last"]] = None,
        timezone: Optional[tz.BaseTzInfo] = None,
        enabled: bool = True,
    ) -> None:
        self._logger = logging.getLogger(self.__module__)
        self._enabled = to_bool(enabled)

        if timezone is None:
            timezone = to_timezone(tzlocal.get_localzone_name())
        self.timezone = timezone

        # TODO: Implement timedelta sanity check
        self.keep = parse_freq(keep)
        self.freq = parse_freq(freq)
        self.resample = parse_freq(resample)

        if aggregate is not None:
            aggregate = aggregate.lower()
            if aggregate not in ["sum", "mean", "min", "max", "last"]:
                raise ConfigurationError(f"Invalid retention aggregation '{aggregate}'")
        self.method = aggregate

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Retention) and self._enabled == other._enabled and self._get_args() == other._get_args()
        )

    def __hash__(self) -> int:
        return hash((self._enabled, *self._get_args()))

    def __repr__(self) -> str:
        return f"{type(self).__name__}(keep={self.keep}, resample={self.resample}, aggregate={self.method})"

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}:\n\t"
            + "\n\t".join(f"{k}={v}" for k, v in self._get_args().items())
            + f"\n\tenabled={self.enabled}"
        )

    # noinspection PyShadowingBuiltins
    def _get_args(self) -> Dict[str, Any]:
        return {
            "keep": self.keep,
            "freq": self.freq,
            "method": self.method,
            "resample": self.resample,
            "timezone": self.timezone,
        }

    @property
    def enabled(self) -> bool:
        return (
            self._enabled
            and self.keep is not None
            and self.freq is not None
            and self.method is not None
            and self.resample is not None
        )

    # noinspection PyProtectedMember, PyTypeChecker, PyShadowingBuiltins
    def aggregate(self, resources: Resources, full: bool = False) -> None:
        if not self.enabled:
            raise RetentionException("Retention aggregation disabled")

        retain = floor_date(pd.Timestamp.now(tz=self.timezone) - to_timedelta(self.keep), freq=self.freq)

        for database, database_resources in resources.groupby(lambda c: c.logger._connector):
            for _, resample_resources in database_resources.groupby(lambda c: c.group):
                start = database.read_first_index(resample_resources)
                if start is not None and start == floor_date(start, freq=self.freq):
                    start += pd.Timedelta(seconds=1)

                end = database.read_last_index(resample_resources)
                if end is not None:
                    end = min(floor_date(end, freq=self.freq), retain)

                if (any(t is None for t in [start, end]) or start >= end) and not full:
                    self._logger.debug(
                        f"Skip aggregating values of resource{'s' if len(resample_resources) > 1 else ''} "
                        + ", ".join([f"'{r.id}'" for r in resample_resources])
                        + " without any new values found"
                    )
                    continue

                if end - start < to_timedelta(self.freq) and not full:
                    self._logger.debug(
                        f"Skip aggregating data of resource{'s' if len(resample_resources) > 1 else ''} "
                        + ", ".join([f"'{r.id}'" for r in resample_resources])
                        + f" to {end.strftime('%d.%m.%Y (%H:%M:%S)')}"
                    )
                    continue

                self._logger.debug(
                    "Start aggregating data"
                    + f" of resource{'s' if len(resample_resources) > 1 else ''} "
                    + ", ".join([f"'{r.id}'" for r in resample_resources])
                    + f" up to {end.strftime('%d.%m.%Y (%H:%M:%S)')}"
                )

                resample_ranges = slice_range(start, end, timezone=self.timezone, freq=self.freq)
                resample_ranges.reverse()
                for resample_start, resample_end in resample_ranges:
                    self._logger.debug(
                        "Start resampling data"
                        + f" of resource{'s' if len(resample_resources) > 1 else ''} "
                        + ", ".join([f"'{r.id}'" for r in resample_resources])
                        + f" from {resample_start.strftime('%d.%m.%Y (%H:%M:%S)')}"
                        + f" to {resample_end.strftime('%d.%m.%Y (%H:%M:%S)')}"
                    )

                    data = database.read(resample_resources, start=resample_start, end=resample_end)
                    if data is None or data.empty:
                        self._logger.debug(
                            "Skipping empty resampling range"
                            + f" of resource{'s' if len(resample_resources) > 1 else ''} "
                            + ", ".join([f"'{r.id}'" for r in resample_resources])
                            + f" from {resample_start.strftime('%d.%m.%Y (%H:%M:%S)')}"
                            + f" to {resample_end.strftime('%d.%m.%Y (%H:%M:%S)')}"
                        )
                        continue

                    # errors = validate(data)
                    # if not errors.empty:
                    #     # TODO: Write detected errors into yaml file or smth similar
                    #     self._logger.warning(
                    #         f"Unable to aggregate data for resource{'s' if len(resample_resources) > 1 else ''} "
                    #         + ", ".join([f"'{r.id}'" for r in resample_resources])
                    #         + f"from {resample_start.strftime('%d.%m.%Y (%H:%M:%S)')} "
                    #         + f"to {resample_end.strftime('%d.%m.%Y (%H:%M:%S)')} "
                    #         + f"with {len(errors)} unprocessed errors"
                    #     )
                    #     print(errors)
                    #     continue

                    resampled_data = resample(data, self.resample, self.method)
                    if hash_data(resampled_data) != hash_data(data):
                        self._logger.debug(
                            f"Starting to resample {len(data)} to {len(resampled_data)} values"
                            + f" of resource{'s' if len(resample_resources) > 1 else ''} "
                            + ", ".join([f"'{r.id}'" for r in resample_resources])
                            + f" from {resample_start.strftime('%d.%m.%Y (%H:%M:%S)')}"
                            + f" to {resample_end.strftime('%d.%m.%Y (%H:%M:%S)')}"
                        )
                        database.delete(resample_resources, start=resample_start, end=resample_end)
                        database.write(resampled_data)

                        self._logger.info(
                            f"Resampled {len(data)} to {len(resampled_data)} values"
                            + f" of resource{'s' if len(resample_resources) > 1 else ''} "
                            + ", ".join([f"'{r.id}'" for r in resample_resources])
                            + f" from {resample_start.strftime('%d.%m.%Y (%H:%M:%S)')}"
                            + f" to {resample_end.strftime('%d.%m.%Y (%H:%M:%S)')}"
                        )
                    elif not not full:
                        self._logger.debug(
                            "Skipping already resampled range"
                            + f" of resource{'s' if len(resample_resources) > 1 else ''} "
                            + ", ".join([f"'{r.id}'" for r in resample_resources])
                            + f" from {resample_start.strftime('%d.%m.%Y (%H:%M:%S)')}"
                            + f" to {resample_end.strftime('%d.%m.%Y (%H:%M:%S)')}"
                        )
                    else:
                        self._logger.debug(
                            "Ending aggregation"
                            + f" of resource{'s' if len(resample_resources) > 1 else ''} "
                            + ", ".join([f"'{r.id}'" for r in resample_resources])
                            + " as data is already resampled"
                            + f" from {resample_start.strftime('%d.%m.%Y (%H:%M:%S)')}"
                            + f" to {resample_end.strftime('%d.%m.%Y (%H:%M:%S)')}"
                        )
                        break


class Retentions(MutableSequence[Retention]):
    def __init__(self, *retentions: Retention) -> None:
        self.__retentions = list(retentions)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(f'{r.keep}' for r in self.__retentions)})"

    def __str__(self) -> str:
        return f"{type(self).__name__}:\n\t" + "\n\t".join(repr(r) for r in self.__retentions)

    def __contains__(self, retention: Retention):
        return self.contains(retention)

    def __getitem__(self, index: int) -> Retention:
        return self.__retentions[index]

    def __iter__(self) -> Iterator[Retention]:
        return iter(self.__retentions)

    def __len__(self) -> int:
        return len(self.__retentions)

    def __setitem__(self, index: int, retention: Retention) -> None:
        self.__retentions[index] = retention

    def __delitem__(self, index: int) -> None:
        del self.__retentions[index]

    def insert(self, index: int, retention: Retention):
        self.__retentions.insert(index, retention)

    def extend(self, retentions: Iterable[Retention], unique: bool = False):
        if unique:
            retentions = [r for r in retentions if r not in self.__retentions]
        self.__retentions.extend(retentions)

    def contains(self, retention: Retention):
        return any(r == retention for r in self.__retentions)

    def sort(self) -> None:
        def order(freq: str) -> int:
            freq_val = "".join(s for s in freq if s.isnumeric())
            freq_val = int(freq_val) if len(freq_val) > 0 else 1
            if freq.endswith("Y"):
                return 31535965 * freq_val
            elif freq.endswith("M"):
                return 2628000 * freq_val
            elif freq.endswith("W"):
                return 604800 * freq_val
            elif freq.endswith("D"):
                return 86400 * freq_val
            elif freq.endswith("h"):
                return 3600 * freq_val
            elif freq.endswith("min"):
                return 60 * freq_val
            elif freq.endswith("s"):
                return 1 * freq_val
            return 0

        self.__retentions = sorted(self.__retentions, key=lambda r: order(r.keep))


class RetentionException(ResourceError):
    """
    Raise if an error occurred while processing retention policy.

    """



# -*- coding: utf-8 -*-
"""
lories.core.configs
~~~~~~~~~~~~~~~~~~~


"""

from .errors import (  # noqa: F401
    ConfigurationError,
    ConfigurationUnavailableError,
)

from .directories import (  # noqa: F401
    Directory,
    Directories,
)

from .configurations import Configurations  # noqa: F401

from .configurator import Configurator  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.simulation.result
~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from collections import OrderedDict
from typing import Any, Dict, Iterator, List, MutableMapping, Optional

from lories.core import Constant


class Result(MutableMapping):
    __map: OrderedDict[str, Any]

    _key: str
    _name: str
    _header: str

    order: int = 100

    summary: Any

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        key: str,
        name: str,
        summary: Any,
        header: str = "Summary",
        order: int = 100,
        **kwargs: Any,
    ) -> None:
        self._key = key
        self._name = name
        self._header = header
        self.__map = OrderedDict(kwargs)

        self.order = order

        self.summary = summary

    def __len__(self) -> int:
        return len(self.__map) + 4

    def __iter__(self) -> Iterator[str]:
        return iter(self._get_attrs())

    def __contains__(self, attr: str) -> bool:
        return attr in self._get_attrs()

    def __getattr__(self, attr: str) -> Any:
        # __getattr__ gets called when the item is not found via __getattribute__
        # To avoid recursion, call __getattribute__ directly to get components dict
        configs = Result.__getattribute__(self, f"_{Result.__name__}__map")
        if attr in configs.keys():
            return configs[attr]
        raise AttributeError(f"'{type(self).__name__}' object has no value '{attr}'")

    def __getitem__(self, key: str) -> Any:
        value = self.get(key)
        if value is not None:
            return value
        raise KeyError(key)

    def __setitem__(self, key: str, value: Any) -> None:
        if key in ["key", "name", "header"]:
            raise KeyError(f"Unable to set result '{key}' attribute")
        elif key == "summary":
            self.summary = value
        else:
            self.__map[key] = value

    def __delitem__(self, key: str) -> None:
        if key in ["key", "name", "header", "summary"]:
            raise KeyError(f"Unable to delete result '{key}' attribute")
        del self.__map[key]

    def get(self, attr: str, default: Optional[Any] = None) -> Any:
        return self._get_vars().get(attr, default)

    def _get_attrs(self) -> List[str]:
        return ["key", "name", "header", "summary", *self.__map.keys()]

    def _get_vars(self) -> Dict[str, Any]:
        return OrderedDict(key=self.key, name=self.name, header=self.header, summary=self.summary, **self.__map)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.key})"

    def __str__(self) -> str:
        return f"{type(self).__name__}:\n\t" + "\n\t".join(f"{k}={v}" for k, v in self._get_vars().items())

    @property
    def key(self) -> str:
        return self._key

    @property
    def name(self) -> str:
        return self._name

    @property
    def header(self) -> str:
        return self._header

    @classmethod
    def from_const(cls, constant: Constant, summary: Any, header: Optional[str] = "Summary", **kwargs) -> Result:
        return cls(constant.key, constant.full_name(unit=True), summary, header=header, **kwargs)



# -*- coding: utf-8 -*-
"""
lories.connector.weather.dwd.brightsky
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import json
from typing import Optional, Tuple

import requests

import numpy as np
import pandas as pd
from lories import ConfigurationError
from lories.components.weather import Weather
from lories.connectors import Connector
from lories.location import Location
from lories.typing import Configurations, Resources, Timestamp


class Brightsky(Connector):
    location: Location
    address: str = "https://api.brightsky.dev/"
    horizon: int = 10

    def __init__(self, context: Weather, location: Location, **kwargs) -> None:
        super().__init__(context, context.configs.get_member("brightsky", defaults={}), **kwargs)
        self.location = location

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)

        self.address = configs.get("address", default=Brightsky.address)
        self.horizon = configs.get_int("horizon", default=Brightsky.horizon)
        if -1 > self.horizon > 10:
            raise ConfigurationError(f"Invalid forecast horizon: {self.horizon}")

    def read(
        self,
        resources: Resources,
        start: Optional[Timestamp] = None,
        end: Optional[Timestamp] = None,
    ) -> pd.DataFrame:
        response, sources = self._request(start, end)
        response_sources = sources.loc[response["source_id"], ["observation_type", "first_record", "last_record"]]
        response_source_columns = ["source_type", "source_first_record", "source_last_record"]
        response_sources.columns = response_source_columns
        response_sources.index = response.index
        response[response_source_columns] = response_sources

        data = []
        for source, source_resources in resources.groupby("source"):
            source_columns = [r.address for r in source_resources if r.address in response.columns]
            source_data = response.loc[
                response["source_type"].isin(s.strip() for s in source.split(",")),
                np.unique(["source_id", "source_first_record", "source_last_record"] + source_columns),
            ]
            if source_data.empty:
                self._logger.warning(f"Unable to read {self._id} channels: {[r.id for r in source_resources]}")
                continue

            source_start = start if start is not None else min(source_data["source_first_record"].unique())
            source_end = end if end is not None else max(source_data["source_last_record"].unique())

            source_data = source_data.rename(columns={r.address: r.id for r in source_resources})

            if source == "forecast":
                source_end = source_start + pd.Timedelta(days=self.horizon)

            elif any(s in ["historical", "current"] for s in source.split(",")):
                if all(t is None for t in [start, end]):
                    source_start = source_end

            data.append(
                source_data.loc[
                    source_start:source_end, [r.id for r in source_resources if r.id in source_data.columns]
                ]
            )
        return pd.concat(data, axis="index")

    # noinspection PyPackageRequirements
    def _request(
        self,
        date: Optional[Timestamp] = None,
        date_last: Optional[Timestamp] = None,
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        if date is None:
            date = pd.Timestamp.now(tz=self.location.timezone)
        if date_last is None:
            date_last = date + pd.Timedelta(days=self.horizon)
        parameters = {
            "date": date.strftime("%Y-%m-%d"),
            "last_date": date_last.strftime("%Y-%m-%d"),
            "lat": self.location.latitude,
            "lon": self.location.longitude,
            "tz": self.location.timezone.zone,
        }
        response = requests.get(self.address + "weather", params=parameters)

        if response.status_code != 200:
            raise requests.HTTPError(
                "Response returned with error " + str(response.status_code) + ": " + response.reason
            )

        response_json = json.loads(response.text)

        sources = pd.DataFrame(response_json["sources"])
        sources = sources.set_index("id")
        sources["first_record"] = pd.to_datetime(sources["first_record"], utc=True)
        sources["last_record"] = pd.to_datetime(sources["last_record"], utc=True)

        data = pd.DataFrame(response_json["weather"])
        data["timestamp"] = pd.to_datetime(data["timestamp"], utc=True)
        data = data.set_index("timestamp").tz_convert(self.location.timezone)
        data.index.name = "timestamp"

        hours = pd.Series(data=data.index, index=data.index).diff().bfill().dt.total_seconds() / 3600.0

        # Convert global horizontal irradiance from kWh/m^2 to W/m^2
        data["solar"] = data["solar"] * hours * 1000

        if data[Weather.CLOUD_COVER].isna().any():
            data[Weather.CLOUD_COVER] = data[Weather.CLOUD_COVER].interpolate(method="linear")

        return data.dropna(how="all", axis="columns"), sources

    def write(self, data: pd.DataFrame) -> None:
        raise NotImplementedError("Brightsky connector does not support writing")



# -*- coding: utf-8 -*-
"""
lories.core.register.context
~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import re
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Collection, Generic, Optional, Sequence, Type

from lories._core._configurations import Configurations  # noqa
from lories._core._data import DataContext, _DataContext, _DataManager  # noqa
from lories._core._registrator import Registrator  # noqa
from lories.core.errors import ResourceError
from lories.core.register._load import _RegistratorLoader
from lories.core.register.registry import Registry
from lories.util import validate_key


# noinspection SpellCheckingInspection, PyProtectedMember
class RegistratorContext(_RegistratorLoader[Registrator], Generic[Registrator]):
    __context: _DataContext

    # noinspection PyUnresolvedReferences
    def __init__(self, context: DataContext, **kwargs) -> None:
        context = self._assert_context(context)
        super().__init__(context._logger, **kwargs)
        self.__context = context

    @classmethod
    def _assert_context(cls, context: DataContext) -> DataContext:
        if context is None or not isinstance(context, _DataManager):
            raise ResourceError(f"Invalid '{cls.__name__}' context: {type(context)}")
        return context

    @property
    def context(self) -> DataContext:
        return self.__context

    # noinspection PyUnresolvedReferences
    @property
    def _executor(self) -> ThreadPoolExecutor:
        return self.__context._executor

    @property
    @abstractmethod
    def _registry(self) -> Registry[Registrator]: ...

    # noinspection PyMethodMayBeStatic
    def get_types(self) -> Collection[str]:
        return self._registry.get_types()

    def get_all(self, *types: str | Type) -> Sequence[Registrator]:
        def _is_type(regitrator) -> bool:
            for _type in types:
                if isinstance(_type, str) and self._registry.has_type(_type):
                    return self._registry.from_type(_type).is_instance(regitrator)
                if isinstance(_type, type) and isinstance(regitrator, _type):
                    return True
            return False

        if len(types) == 0:
            return list(self.values())
        return self.filter(_is_type)

    def _create(
        self,
        context: RegistratorContext | Registrator,
        configs: Configurations,
        **kwargs: Any,
    ) -> Registrator:
        registration_class = self._get_registator_class()
        registrator_member = configs.get_member(registration_class.TYPE, ensure_exists=True)

        if "key" not in registrator_member:
            if "key" in configs:
                registration_key = configs.get("key")
                del configs["key"]
            else:
                registration_key = "_".join(re.split(r"[^\w-]", configs.key))
            registrator_member["key"] = validate_key(registration_key)
            registrator_member.move_to_top("key")

        if "name" not in registrator_member and "name" in configs:
            registration_name = configs.get("name")
            registrator_member["name"] = registration_name
            registrator_member.move_to_top("name")
            del configs["name"]

        registration_type = re.split(r"[^\w_-]", configs.key)[0]
        if "type" in registrator_member:
            registration_type = validate_key(registrator_member.get("type"))
        elif "type" in configs:
            _registration_type = validate_key(configs.get("type"))
            if self._registry.has_type(_registration_type) or not self._registry.has_type(registration_type):
                registration_type = _registration_type
                del configs["type"]

        registration = self._registry.from_type(registration_type)
        if "type" not in registrator_member:
            registrator_member["type"] = registration.key
            registrator_member.move_to_top("type")
        return registration.initialize(context, configs, **kwargs)

    # noinspection PyUnresolvedReferences
    def _load_registrators_configs(self) -> Configurations:
        _type = self._get_registrators_type()
        return self.__context.configs.get_member(_type, ensure_exists=True)

    def load(self, configs: Optional[Configurations] = None, **kwargs) -> Sequence[Registrator]:
        _class = self._get_registator_class()
        if configs is None:
            configs = self._load_registrators_configs()
        return self._load(self, configs, includes=_class.INCLUDES, **kwargs)



# -*- coding: utf-8 -*-
"""
lories.application.view.interface
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
import os
import shutil
from importlib import resources
from pathlib import Path
from typing import Optional

import dash
from dash import Dash, dcc, html
from dash_bootstrap_components import themes

from lories.application import Application
from lories.application.interface import Interface, register_interface_type
from lories.application.view import LoginPage, PageFooter, PageHeader, View
from lories.typing import Configurations

logging.getLogger("werkzeug").setLevel(logging.WARNING)


# noinspection PyProtectedMember
@register_interface_type("dash")
class ViewInterface(Interface, Dash):
    _proxy: Optional[str] = None
    _host: str
    _port: int

    def __init__(self, context: Application, configs: Configurations) -> None:
        def get_custom_path(key: str, default: Optional[str] = None) -> str:
            if "pages" in configs:
                custom_path = configs[key]
            else:
                custom_path = default
            if custom_path is None:
                return ""
            if os.path.isabs(custom_path):
                custom_path = Path(custom_path)
            else:
                custom_path = Path(configs.dirs.data, custom_path)
            if custom_path is not None and not custom_path.exists():
                custom_path.mkdir(exist_ok=True)
            return str(custom_path)

        view_path = resources.files("lories.application.view")
        pages_path = get_custom_path("pages")

        assets_default = str(view_path.joinpath("assets"))
        assets_path = get_custom_path("assets", default=assets_default)
        if assets_path != assets_default:

            def copy_assets(src, dest):
                dest = Path(dest)
                if dest.is_dir():
                    return
                if dest.exists() or dest.suffix not in [".ico", ".png", ".jpg", ".jpeg", ".css"]:
                    return
                shutil.copy2(src, dest)

            shutil.copytree(
                assets_default,
                assets_path,
                dirs_exist_ok=True,
                copy_function=copy_assets,
            )

        super().__init__(
            name=context.name,
            title=context.name,
            context=context,
            configs=configs,
            external_stylesheets=[themes.BOOTSTRAP],
            assets_folder=assets_path,
            pages_folder=pages_path,
            use_pages=True,
            server=True,  # TODO: Replace this with local Flask server, to create custom REST API ?
        )
        theme_defaults = {
            "name": context.name,
            "logo": os.path.join(assets_path, "logo.png"),
        }
        theme = configs.get_member("theme", defaults=theme_defaults)

        header = PageHeader(**theme)
        footer = PageFooter()

        self.view = View(context.id, header, footer)

        login = configs.get_member("login", defaults={"enabled": False})
        if login.enabled:
            login_page = LoginPage(self, context, configs)
            self.view.append(login_page)

    # noinspection PyUnresolvedReferences
    def configure(self, configs: Configurations) -> None:
        super().configure(configs)
        self._proxy = configs.get("proxy", default=None)
        self._host = configs.get("host", default="127.0.0.1")
        self._port = configs.get_int("port", default=8050)

        self.view.create_pages(self.context.components)
        self.view.create_layout(self.view.layout)
        self.view.register()
        self.layout = self.create_layout

    def start(self) -> None:
        self.run(
            host=self._host,
            port=self._port,
            proxy=self._proxy,
            debug=self._logger.getEffectiveLevel() <= logging.DEBUG,
        )

    # noinspection PyUnresolvedReferences
    def create_layout(self) -> html.Div:
        return html.Div(
            id=f"{self.context.id}",
            children=[
                self.view.header.navbar,
                dash.page_container,
                dcc.Interval(
                    id="view-update",
                    interval=1000,
                ),
            ],
        )



# -*- coding: utf-8 -*-
"""
lories._data._channels
~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable
from typing import Iterable, TypeVar, Union

import pandas as pd
from lories._core._channel import ChannelState, _Channel
from lories._core._resources import _Resources

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


class _Channels(_Resources[_Channel]):
    TYPE: str = "channels"

    @abstractmethod
    def register(
        self,
        function: Callable[[pd.DataFrame], None],
        how: Literal["any", "all"] = "any",
        unique: bool = False,
    ) -> None: ...

    @abstractmethod
    def from_logger(self) -> Channels: ...

    @abstractmethod
    def to_frame(self, unique: bool = False, states: bool = False) -> pd.DataFrame: ...

    @abstractmethod
    # noinspection PyProtectedMember
    def set_frame(self, data: pd.DataFrame) -> None: ...

    @abstractmethod
    def set_state(self, state: ChannelState) -> None: ...


Channels = TypeVar("Channels", bound=_Channels)

ChannelsArgument = TypeVar(
    "ChannelsArgument",
    bound=Union[_Channel, _Channels, Iterable[_Channel], Iterable[str], str],
)



# -*- coding: utf-8 -*-
"""
lories.data.context
~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from collections.abc import Mapping
from copy import deepcopy
from typing import Any, Callable, Collection, Iterable, List, Optional, Tuple

import numpy as np
import pandas as pd
from lories._core import _DataContext  # noqa
from lories.core.configs import ConfigurationError, Configurations, Directories
from lories.core.errors import ResourceError
from lories.core.typing import ChannelsArgument, ContextArgument
from lories.data.channels import Channel, Channels
from lories.util import update_recursive, validate_key


# noinspection PyAbstractClass, PyProtectedMember
class DataContext(_DataContext):
    def _load(
        self,
        context: ContextArgument,
        configs: Configurations,
        sort: bool = True,
    ) -> Collection[Channel]:
        channels = []

        defaults = {}
        configs = configs.copy()
        if configs.has_member(self.TYPE):
            data = configs.get_member(self.TYPE)
            update_recursive(defaults, Channel._build_defaults(configs))
            if data.has_member(Channels.TYPE):
                channels.extend(self._load_from_members(context, data.get_member(Channels.TYPE), defaults))
        channels.extend(self._load_from_file(context, configs.dirs, f"{Channels.TYPE}.conf", defaults=defaults))

        if sort:
            self.sort()
        return channels

    # noinspection PyShadowingBuiltins
    def _load_from_configs(
        self,
        context: ContextArgument,
        key: str,
        **configs: Any,
    ) -> Channel:
        id = Channel._build_id(key=key, context=context)
        if "type" not in configs:
            raise ConfigurationError("Missing 'type' for channel: " + id)

        if self._contains(id):
            self._update(id=id, key=key, **configs)
            return self._get(id)

        channel = self._create(id=id, key=key, **configs)
        self._add(channel)
        return channel

    def _load_from_members(
        self,
        context: ContextArgument,
        configs: Configurations,
        defaults: Optional[Mapping[str, Any]] = None,
    ) -> Collection[Channel]:
        channels = []
        if defaults is None:
            defaults = {}
        update_recursive(defaults, Channel._build_defaults(configs))

        for channel_key in [i for i in configs.keys() if i not in defaults]:
            channel_configs = update_recursive(deepcopy(defaults), configs.get_member(channel_key))
            channel_key = validate_key(channel_configs.pop("key", channel_key))
            channels.append(self._load_from_configs(context, channel_key, **channel_configs))
        return channels

    def _load_from_file(
        self,
        context: ContextArgument,
        configs_dirs: Directories,
        configs_file: str,
        defaults: Optional[Mapping[str, Any]] = None,
    ) -> Collection[Channel]:
        channels = []
        if configs_dirs.conf.joinpath(configs_file).is_file():
            configs = Configurations.load(configs_file, **configs_dirs.to_dict())
            channels.extend(self._load_from_members(context, configs, defaults))
        return channels

    # noinspection PyShadowingBuiltins
    def _set(self, id: str, channel: Channel) -> None:
        if not isinstance(channel, Channel):
            raise ResourceError(f"Invalid channel type: {type(channel)}")

        # TODO: connector sanity check
        super()._set(id, channel)

    # noinspection PyShadowingBuiltins, PyProtectedMember, PyArgumentList
    def _update(self, id: str, key: str, **configs: Any) -> None:
        channel = self._get(id)
        channel._update(**configs)

    @property
    def channels(self) -> Channels:
        return Channels(self.values())

    def _filter_by_args(self, channels: Optional[ChannelsArgument]) -> Channels:
        if channels is None:
            return self.channels
        _channels = []

        def append(_channel: Channel | str) -> None:
            if isinstance(_channel, str):
                if _channel in self:
                    _channels.append(self[_channel])
            elif isinstance(_channel, Channel):
                _channels.append(_channel)
            else:
                raise ResourceError(f"Invalid '{type(_channel)}' channel: {_channel}")

        if isinstance(channels, str) or isinstance(channels, Channel):
            append(channels)
        elif isinstance(channels, Iterable):
            for channel in channels:
                append(channel)
        else:
            raise ResourceError(f"Invalid '{type(channels)}' channels: {channels}")

        return Channels(_channels)

    # noinspection PyShadowingBuiltins
    def filter(self, *filters: Optional[Callable[[Channel], bool]]) -> Channels:
        return Channels(super().filter(*filters))

    # noinspection SpellCheckingInspection
    def groupby(self, by: str) -> List[Tuple[Any, Channels]]:
        groups = []
        for group_by in np.unique([getattr(c, by) for c in self.values()]):
            groups.append((group_by, self.filter(lambda c: getattr(c, by) == group_by)))
        return groups

    def to_frame(self, **kwargs) -> pd.DataFrame:
        return self.channels.to_frame(**kwargs)



# -*- coding: utf-8 -*-
"""
lories.application.view.pages.layout.card
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from collections.abc import MutableSequence
from typing import Any, Dict, Iterable, List, Optional

import dash_bootstrap_components as dbc
from dash import html


class PageCardItem(dbc.ListGroupItem):
    focus: bool

    # noinspection PyPep8Naming
    def __init__(
        self,
        children: Any,
        class_name: Optional[str] = None,
        className: Optional[str] = None,
        focus: bool = False,
        **kwargs,
    ) -> None:
        if class_name is None or len(class_name) == 0:
            class_name = className
        if class_name is None or len(class_name) == 0:
            class_name = "card-item"
        else:
            class_name = f"{class_name} card-item"

        self.focus = focus
        if focus:
            class_name = class_name.replace("card-item", "card-focus")

        super().__init__(children=children, class_name=class_name, **kwargs)


class PageCard(dbc.Card, MutableSequence[PageCardItem]):
    items = List[PageCardItem]

    def __init__(self, horizontal: bool = True, style: Dict[str, Any] = None, **kwargs) -> None:
        self.items = []
        self._body = dbc.CardBody([dbc.ListGroup(self.items, horizontal=horizontal, flush=True, style=style)])
        super().__init__(children=[self._body], **kwargs)

    def __len__(self) -> int:
        return len(self.items)

    def __getitem__(self, index: int) -> PageCardItem:
        return self.items[index]

    def __setitem__(self, index: int, item: PageCardItem) -> None:
        self.items[index] = item

    def __delitem__(self, index: int) -> None:
        del self.items[index]

    def insert(self, index: int, item: PageCardItem) -> None:
        self.items.insert(index, item)

    def append(self, *children: Any, focus: bool = False) -> None:
        self.items.append(PageCardItem(list(children), focus=focus))

    def extend(self, items: Iterable[PageCardItem] | PageCard) -> None:
        if isinstance(items, PageCard):
            items = items.items
        self.items.extend(items)

    def has_items(self) -> bool:
        return len(self.items) > 0

    def add_title(self, title: str) -> None:
        self._body.children.insert(0, html.H4(title, className="card-title"))

    def add_header(self, *children: Any, **kwargs) -> None:
        children = list(children)
        self.children.insert(0, dbc.CardHeader(children, **kwargs))

    def add_footer(self, *children: Any, href: Optional[str] = None, **kwargs) -> None:
        children = list(children)
        if href is not None and len(href) > 0:
            children.append(dbc.Button("More", class_name="card-button", href=href))
        self.children.append(dbc.CardFooter(children, **kwargs))



# -*- coding: utf-8 -*-
"""
lories.data.replication
~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
from collections.abc import Sequence
from typing import Any, Dict, Iterator, List, Mapping, Optional

import tzlocal

import pandas as pd
import pytz as tz
from lories import ConfigurationError, Resource, ResourceError, Resources
from lories.connectors import Database
from lories.util import floor_date, parse_freq, slice_range, to_bool, to_timedelta, to_timezone

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


class Replication:
    TYPE: str = "replication"

    _enabled: bool = False

    database: Database

    method: Literal["push", "pull"]
    slice: str = "D"
    freq: str = "D"
    timezone: tz.BaseTzInfo

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        database: Optional[Database],
        timezone: Optional[tz.BaseTzInfo] = None,
        method: Literal["push", "pull"] = "push",
        slice: str = "D",
        freq: str = "D",
        enabled: bool = True,
    ) -> None:
        self._logger = logging.getLogger(self.__module__)
        self._enabled = to_bool(enabled)

        self.database = self._assert_database(database)

        if timezone is None:
            timezone = to_timezone(tzlocal.get_localzone_name())
        self.timezone = timezone

        if method not in ["push", "pull"]:
            raise ConfigurationError(f"Invalid replication method '{method}'")
        self.method = method
        self.freq = parse_freq(freq)
        self.slice = parse_freq(slice)

    @classmethod
    def _assert_database(cls, database):
        if database is None:
            return None
        if not isinstance(database, Database):
            raise ResourceError(database, f"Invalid database: {None if database is None else type(database)}")
        return database

    # noinspection PyShadowingBuiltins
    def _is_mapped(
        self,
        database: Optional[str],
        timezone: Optional[tz.BaseTzInfo] = None,
        method: Literal["push", "pull"] = "push",
        slice: str = "D",
        freq: str = "D",
        enabled: bool = True,
    ) -> bool:
        return (
            self.database.key == database
            and self.timezone == to_timezone(timezone)
            and self.method == method
            and self.freq == parse_freq(freq)
            and self.slice == parse_freq(slice)
            and self._enabled == to_bool(enabled)
        )

    def _is_equal(self, other: Replication) -> bool:
        return (
            self._enabled == other._enabled
            and self.database.id == other.database.id
            and self._get_args() == other._get_args()
        )

    def __eq__(self, other: Any) -> bool:
        return self._is_equal(other)

    def __hash__(self) -> int:
        return hash((self.database, self._enabled, *self._get_args()))

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.id})"

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}:\n\tid={self.id}\n\t"
            + "\n\t".join(f"{k}={v}" for k, v in self._get_args().items())
            + f"\n\tenabled={self.enabled}"
        )

    @property
    def id(self) -> Optional[str]:
        return self.database.id if self.database is not None else None

    @property
    def key(self) -> str:
        return self.database.key if self.database is not None else None

    # noinspection PyShadowingBuiltins
    def _get_args(self) -> Dict[str, Any]:
        return {
            "method": self.method,
            "slice": self.slice,
            "freq": self.freq,
            "timezone": self.timezone,
        }

    @property
    def enabled(self) -> bool:
        return self._enabled and self.database is not None and self.database.is_enabled()

    # noinspection PyProtectedMember
    def replicate(self, resources: Resources, full: bool = False, force: bool = False) -> None:
        if not self.enabled:
            raise ReplicationException(self.database, "Replication disabled")
        if not self.database.is_connected():
            raise ReplicationException(self.database, f"Replication database '{self.database.id}' not connected")
        kwargs = self._get_args()
        kwargs["full"] = full
        kwargs["force"] = force
        method = kwargs.pop("method")

        for logger, logger_resources in resources.groupby(lambda c: c.logger._connector):
            for _, group_resources in logger_resources.groupby(lambda c: c.group):
                try:
                    if method == "push":
                        replicate(logger, self.database, group_resources, **kwargs)
                    elif method == "pull":
                        replicate(self.database, logger, group_resources, **kwargs)

                except ReplicationException as e:
                    self._logger.error(f"Replication failed because: {e}")


class ReplicationException(ResourceError):
    """
    Raise if an error occurred while replicating.

    """


class Replications(Sequence[Replication]):
    _replications: List[Replication]

    # noinspection PyProtectedMember, PyShadowingNames
    def build(self, databases, resource: Resource, **configs) -> Optional[Replication]:
        resource_configs = resource.get(Replication.TYPE, default=None)
        if resource_configs is None or "database" not in resource_configs:
            return None
        if isinstance(resource_configs, str):
            resource_configs = {"database": resource_configs}
        elif not isinstance(resource_configs, Mapping):
            raise ConfigurationError("Invalid resource replication database: " + str(resource_configs))

        replication = None
        for _replication in self:
            if _replication._is_mapped(**resource_configs):
                replication = _replication
                break
        if replication is None:
            database = None
            database_id = resource_configs.pop("database")
            if database_id is not None:
                database_path = resource.id.split(".")
                for i in reversed(range(len(database_path))):
                    _database_id = ".".join([*database_path[:i], database_id])
                    if _database_id in databases.keys():
                        database = databases.get(_database_id, None)
                        break

            configs.update(resource_configs)
            replication = Replication(database, **configs)

            self._replications.append(replication)
        return replication

    def __init__(self, resources=()) -> None:
        self._logger = logging.getLogger(type(self).__module__)
        self._replications = [*resources]

    # noinspection PyProtectedMember
    def __contains__(self, _object: Mapping[str, Any] | Replication) -> bool:
        if isinstance(_object, Replication):
            return _object in self._replications
        if isinstance(_object, Mapping):
            return any(r._is_mapped(*_object) for r in self._replications)
        return False

    def __getitem__(self, _index: int):
        return self._replications[_index]

    def __iter__(self) -> Iterator[Replication]:
        return iter(self._replications)

    def __len__(self) -> int:
        return len(self._replications)

    def __add__(self, _object):
        return type(self)([*self, *_object])

    def append(self, replication: Replication) -> None:
        self._replications.append(replication)


# noinspection PyProtectedMember, PyUnresolvedReferences, PyTypeChecker, PyShadowingBuiltins
def replicate(
    source: Database,
    target: Database,
    resources: Resources,
    timezone: Optional[tz.BaseTzInfo] = None,
    slice: str = "D",
    freq: str = "D",
    full: bool = True,
    force: bool = False,
) -> None:
    if source is None or target is None or len(resources) == 0:
        return

    logger = logging.getLogger(Replication.__module__)
    logger.debug(
        f"Starting to replicate data of resource{'s' if len(resources) > 1 else ''} "
        + ", ".join([f"'{r.id}'" for r in resources])
    )
    target_empty = False

    if timezone is None:
        timezone = to_timezone(tzlocal.get_localzone_name())
    now = pd.Timestamp.now(tz=timezone)
    end = source.read_last_index(resources)
    if end is None:
        end = now
    if not full:
        end = floor_date(end, freq=freq)

    start = target.read_last_index(resources) if not full else None
    if start is None:
        start = source.read_first_index(resources)
        target_empty = True

    if any(t is None for t in [start, end]) or start >= end:
        logger.debug(
            f"Skip copying values of resource{'s' if len(resources) > 1 else ''} "
            + ", ".join([f"'{r.id}'" for r in resources])
            + " without any values found"
        )
        return

    if not target_empty:
        # Validate prior step, before continuing
        prior_end = floor_date(start if start <= now else now, timezone=timezone, freq=freq)
        prior_start = prior_end - to_timedelta(freq) + pd.Timedelta(seconds=1)
        replicate_range(source, target, resources, prior_start, prior_end, force=force)

    if start + to_timedelta(slice) < end:
        for slice_start, slice_end in slice_range(start, end, timezone=timezone, freq=slice):
            replicate_range(source, target, resources, slice_start, slice_end, force=force)
    else:
        replicate_range(source, target, resources, start, end, force=force)


def replicate_range(
    source: Database,
    target: Database,
    resources: Resources,
    start: pd.Timestamp,
    end: pd.Timestamp,
    force: bool = False,
) -> None:
    logger = logging.getLogger(Replication.__module__)
    logger.debug(
        f"Start copying data of resource{'s' if len(resources) > 1 else ''} "
        + ", ".join([f"'{r.id}'" for r in resources])
        + f" from {start.strftime('%d.%m.%Y (%H:%M:%S)')}"
        + f" to {end.strftime('%d.%m.%Y (%H:%M:%S)')}"
    )

    source_checksum = source.hash(resources, start, end)
    if source_checksum is None:
        logger.debug(
            f"Skipping time slice without database data for resource{'s' if len(resources) > 1 else ''} "
            + ", ".join([f"'{r.id}'" for r in resources]),
        )
        return

    target_checksum = target.hash(resources, start, end)
    if target_checksum == source_checksum:
        logger.debug(
            f"Skipping time slice without changed data for resource{'s' if len(resources) > 1 else ''} "
            + ", ".join([f"'{r.id}'" for r in resources])
        )
        return

    data = source.read(resources, start=start, end=end)
    if data is None or data.empty:  # not source.exists(resources, start=start, end=end):
        logger.debug(
            f"Skipping time slice without new data for resource{'s' if len(resources) > 1 else ''} "
            + ", ".join([f"'{r.id}'" for r in resources]),
        )
        return

    logger.debug(
        f"Copying {len(data)} values of resource{'s' if len(resources) > 1 else ''} "
        + ", ".join([f"'{r.id}'" for r in resources])
        + f" from {start.strftime('%d.%m.%Y (%H:%M:%S)')}"
        + f" to {end.strftime('%d.%m.%Y (%H:%M:%S)')}"
    )

    target.write(data)
    target_checksum = target.hash(resources, start, end)
    if target_checksum != source_checksum:
        if force:
            target.delete(resources, start, end)
            target.write(data)
            return

        logger.error(
            f"Mismatching for {len(data)} values of resource{'s' if len(resources) > 1 else ''} "
            + ",".join([f"'{r.id}'" for r in resources])
            + f" with checksum '{source_checksum}' against target checksum '{target_checksum}'"
            + f" from {start.strftime('%d.%m.%Y (%H:%M:%S)')}"
            + f" to {end.strftime('%d.%m.%Y (%H:%M:%S)')}"
        )
        raise ReplicationException("Checksum mismatch while synchronizing")

    logger.info(
        f"Replicated {len(data)} values of resource{'s' if len(resources) > 1 else ''} "
        + ", ".join([f"'{r.id}'" for r in resources])
        + f" from {start.strftime('%d.%m.%Y (%H:%M:%S)')}"
        + f" to {end.strftime('%d.%m.%Y (%H:%M:%S)')}"
    )



# -*- coding: utf-8 -*-
"""
lories.connectors.sql.columns.datetime
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import datetime as dt
from typing import Any, Optional, TypeVar

from sqlalchemy.sql import func
from sqlalchemy.types import DATE, DATETIME, TIME, TIMESTAMP

import pandas as pd
import pytz as tz
from lories.connectors.sql.columns import Column, ColumnType
from lories.core.errors import ResourceError

DatetimeType = TypeVar("DatetimeType", pd.DatetimeIndex, pd.Series, pd.Timestamp, dt.datetime, dt.date, dt.time)


class DatetimeColumn(Column):
    inherit_cache: bool = True

    timezone: tz.BaseTzInfo

    # noinspection PyShadowingBuiltins, PyTypeChecker
    def __init__(
        self,
        name: str,
        type: ColumnType,
        *args,
        default: Optional[Any] = None,
        nullable: bool = True,
        timezone: tz.BaseTzInfo = tz.UTC,
        **kwargs,
    ) -> None:
        if not is_datetime(type):
            raise ResourceError(f"Invalid data type for DatetimeColumn: {type}")

        if default is None and not nullable:
            if type == TIMESTAMP:
                default = func.current_timestamp()
            elif type == DATETIME:
                default = func.now()
            elif type == DATE:
                default = func.current_date()
            elif type == TIME:
                default = func.current_time()
        super().__init__(
            name,
            type(timezone),
            *args,
            default=default,
            nullable=nullable,
            **kwargs,
        )
        self.timezone = timezone

    # noinspection PyUnresolvedReferences
    def validate(self, data: Any) -> Any:
        if not is_datetime(self.type):
            return data
        if (
            data is None
            or not isinstance(data, (pd.Series, pd.DatetimeIndex, dt.datetime, dt.date, dt.time))
            or (
                isinstance(data, pd.Series)
                and not data.map(lambda i: isinstance(i, (dt.datetime, dt.date, dt.time))).all()
            )
        ):
            raise ResourceError(f"Unable to prepare datetime index from '{type(data)}' data: {data}")

        if isinstance(data, (pd.DatetimeIndex, pd.Timestamp)):
            data = data.tz_convert(self.timezone).tz_localize(None)

        elif isinstance(data, pd.Series):
            data = data.dt.tz_convert(self.timezone).dt.tz_localize(None)
        else:
            data = data.astimezone(self.timezone).replace(tzinfo=None)

        if type == DATE:
            if isinstance(data, (pd.DatetimeIndex, pd.Series)):
                data = data.date
            else:
                data = data.date()
        elif type == TIME:
            if isinstance(data, (pd.DatetimeIndex, pd.Series)):
                data = data.time
            else:
                data = data.time()

        return super().validate(data)


# noinspection PyShadowingBuiltins, SpellCheckingInspection
def is_datetime(type: ColumnType) -> bool:
    datetimes = (
        TIME,
        DATE,
        DATETIME,
        TIMESTAMP,
    )
    return type in datetimes or isinstance(type, datetimes)



# -*- coding: utf-8 -*-
"""
lories.data.util
~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import hashlib
import re
from copy import copy, deepcopy
from typing import Optional, Tuple

import numpy as np
import pandas as pd
import pytz as tz
from pandas.tseries.frequencies import to_offset

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


# noinspection PyShadowingBuiltins
def hash_data(
    data: pd.DataFrame,
    method: Literal["MD5", "SHA1", "SHA256", "SHA512"] = "MD5",
    encoding: str = "UTF-8",
) -> str:
    index_column = data.index.name if data.index is not None else "index"
    data_columns = data.columns
    data = deepcopy(data)

    for column in data.select_dtypes(include=["datetime64", "datetimetz"]).columns:
        data[column] = data[column].dt.tz_convert(tz.UTC).view(np.int64) // 10**9

    data[data.index.name] = data.index.tz_convert(tz.UTC).view(np.int64) // 10**9
    data = data[[index_column, *data_columns]]

    csv = data.to_csv(index=False, header=False, sep=",", decimal=".", float_format="%.10g")
    csv = ",".join(re.sub(r",,+", ",", line).strip(",") for line in csv.splitlines())
    return hash_value(csv, method, encoding)


def hash_value(
    value: str,
    method: Literal["MD5", "SHA1", "SHA256", "SHA512"],
    encoding: str = "UTF-8",
) -> str:
    method = method.lower()
    if method == "md5":
        return hashlib.md5(value.encode(encoding)).hexdigest()
    if method == "sha1":
        return hashlib.sha1(value.encode(encoding)).hexdigest()
    if method == "sha256":
        return hashlib.sha256(value.encode(encoding)).hexdigest()
    if method == "sha512":
        return hashlib.sha512(value.encode(encoding)).hexdigest()
    raise ValueError(f"Invalid checksum method '{method}'")


# noinspection PyUnresolvedReferences
def resample(
    data: pd.DataFrame | pd.Series,
    freq: str,
    func: Literal["sum", "mean", "min", "max", "last"],
    offset: Optional[pd.Timedelta] = None,
) -> pd.DataFrame | pd.Series:
    if func not in ["sum", "mean", "min", "max", "last"]:
        raise ValueError(f"Invalid resampling function '{func}'")
    freq = to_offset(freq)

    index = copy(data.index)
    index_freq = index.freq
    if index_freq is None and len(index) > 2:
        index_freq = to_offset(pd.infer_freq(index))
    if index_freq is None or index_freq < freq:
        resampled = data.resample(freq, closed="right", offset=offset)
        if func == "sum":
            data = resampled.sum()
        elif func == "mean":
            data = resampled.mean()
        elif func == "min":
            data = resampled.min()
        elif func == "max":
            data = resampled.max()
        elif func == "last":
            data = resampled.last()
        data.index += freq

    data.dropna(axis="columns", how="all", inplace=True)
    data.dropna(axis="index", how="all", inplace=True)
    data.index.name = index.name
    return data


def scale_power(name: str, power: float) -> Tuple[str, float]:
    if power >= 1e7:
        power = round(power / 1e6, 2)
        name = name.replace("W", "MW")
    elif power >= 1e4:
        power = round(power / 1e3, 2)
        name = name.replace("W", "kW")
    else:
        power = round(power, 2)
    return name, power


def scale_energy(name: str, energy: float) -> Tuple[str, float]:
    if energy >= 1e7:
        energy = round(energy / 1e6, 2)
        name = name.replace("kWh", "GWh")
    elif energy >= 1e4:
        energy = round(energy / 1e3, 2)
        name = name.replace("kWh", "MWh")
    else:
        energy = round(energy, 2)
    return name, energy


def derive_by_hours(data: pd.Series) -> pd.Series:
    """
    Derive a data series by hours.

    Parameters
    ----------
    data : pandas.Series
        Series with the data to be derived

    Returns
    ----------
    fixed: pandas.Series
        Series with the derived data

    """
    delta_value = data.iloc[:].astype("float64").diff()

    delta_index = pd.Series(delta_value.index, index=delta_value.index)
    delta_index = (delta_index - delta_index.shift(1)) / np.timedelta64(1, "h")

    return pd.Series(delta_value / delta_index, index=data.index).dropna()



# -*- coding: utf-8 -*-
"""
lories.connectors.modbus
~~~~~~~~~~~~~~~~~~~~~~~~


"""

from .register import ModbusRegister  # noqa: F401

from . import client  # noqa: F401
from .client import ModbusClient  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories.connectors.modbus.client
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from typing import Mapping

from pymodbus import FramerType, ModbusException
from pymodbus.client import ModbusBaseSyncClient, ModbusSerialClient, ModbusTcpClient, ModbusUdpClient

import pandas as pd
import pytz as tz
from lories._core import ChannelState  # noqa
from lories.connectors import ConnectionError, Connector, ConnectorError, register_connector_type
from lories.connectors.modbus import ModbusRegister
from lories.core.configs import ConfigurationError
from lories.typing import Configurations, Resources

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


@register_connector_type("modbus")
class ModbusClient(Connector):
    __client: ModbusBaseSyncClient
    __registers: Mapping[str, ModbusRegister]

    _endian: Literal["big", "little"]

    # noinspection SpellCheckingInspection
    def configure(self, configs: Configurations) -> None:
        super().configure(configs)
        _endian = configs.get("endian", default="big").lower()
        if _endian not in ["big", "little"]:
            raise ConnectorError(self, f"Invalid modbus word order '{_endian}'")
        self._endian = _endian

        timeout = configs.get_int("timeout", default=3)
        retries = configs.get_int("retries", default=3)

        protocol = configs.get("protocol").lower()
        if protocol == "tcp":
            self.__client = ModbusTcpClient(
                host=configs.get("host"),
                port=configs.get_int("port", 502),
                framer=FramerType.SOCKET,
                timeout=timeout,
                retries=retries,
                # source_address=("localhost", 0),
            )
        elif protocol == "udp":
            self.__client = ModbusUdpClient(
                host=configs.get("host"),
                port=configs.get_int("port", 502),
                framer=FramerType.SOCKET,
                timeout=timeout,
                retries=retries,
                # source_address=None,
            )
        elif protocol in ["rtu", "serial"]:
            self.__client = ModbusSerialClient(
                port=configs.get("port"),
                framer=FramerType.RTU,
                timeout=timeout,
                retries=retries,
                baudrate=configs.get_int("baudrate"),
                bytesize=configs.get_int("bytesize", default=8),
                stopbits=configs.get_int("stopbits", default=1),
                parity=configs.get("parity", default="N"),
                # handle_local_echo=False,
            )
        else:
            raise ConnectorError(self, f"Unknown modbus protocol type '{protocol}'")

    # noinspection PyUnresolvedReferences
    def is_connected(self) -> bool:
        return self.__client.connected

    def connect(self, resources: Resources) -> None:
        super().connect(resources)
        try:
            self._logger.info(f"Connecting to '{self.__client}'")
            self.__client.connect()
            self.__registers = {r.id: ModbusRegister.from_resource(r) for r in resources}

        except ModbusException as e:
            self._logger.warning(f"Error connecting to '{self.__client}': {e}")
            raise ConnectionError(self, e)
        except IOError as e:
            raise ConnectorError(self, e)

    def disconnect(self) -> None:
        super().disconnect()
        self.__client.close()

    # noinspection PyTypeChecker, PyShadowingBuiltins
    def read(self, resources: Resources) -> pd.DataFrame:
        timestamp = pd.Timestamp.now(tz.UTC).floor(freq="s")
        data = pd.DataFrame(index=[timestamp], columns=resources.ids)
        try:
            for device, device_resources in resources.groupby("device"):
                if device is None:
                    device = 1

                # TODO: Implement reading adjacent blocks of registers of same device ID
                for resource in device_resources:
                    try:
                        register = self.__registers[resource.id]
                        function = getattr(self.__client, f"read_{register.function}s")
                        result = function(register.address, count=register.length, slave=device)
                        if result.isError():
                            data.at[timestamp, resource.id] = ChannelState.UNKNOWN_ERROR
                            self._logger.warning(f"Error reading register '{resource.id}'")
                            continue

                        value = self.__client.convert_from_registers(
                            result.registers, register.type, word_order=self._endian
                        )
                        data.at[timestamp, resource.id] = value

                        self._logger.debug(f"Read {register.type} value of register {register.address}: {value}")

                    except ConfigurationError as e:
                        data.at[timestamp, resource.id] = ChannelState.ARGUMENT_SYNTAX_ERROR
                        self._logger.warning(f"Invalid register configuration for resource '{resource.id}': {e}")
                        continue
                    except KeyError:
                        data.at[timestamp, resource.id] = ChannelState.NOT_AVAILABLE
                        continue
            return data

        except ModbusException as e:
            raise ConnectionError(self, e)
        except IOError as e:
            raise ConnectorError(self, e)

    def write(self, data: pd.DataFrame) -> None:
        try:
            for device, device_channels in self.channels.groupby("device"):
                if device is None:
                    device = 1

                for channel in device_channels:
                    if channel.id not in data.columns:
                        continue
                    channel_data = data.loc[:, channel.id].dropna(axis="index", how="all")
                    if channel_data.empty:
                        continue
                    register = self.__registers[channel.id]
                    try:
                        values = self.__client.convert_to_registers(
                            channel_data.iloc[-1], register.type, word_order=self._endian
                        )
                        self.__client.write_registers(register.address, values, slave=device)

                    except ConfigurationError as e:
                        self._logger.warning(f"Invalid register configuration for channel '{channel.id}': {e}")
                        continue

        except ModbusException as e:
            raise ConnectionError(self, e)
        except IOError as e:
            raise ConnectorError(self, e)



# -*- coding: utf-8 -*-
"""
lories.connectors.sql
~~~~~~~~~~~~~~~~~~~~~


"""

from .columns import (  # noqa: F401
    Column,
    DatetimeColumn,
    UnixTimestampColumn,
    SurrogateKeyColumn,
)

from .index import DatetimeIndexType  # noqa: F401

from .table import Table  # noqa: F401

from .schema import Schema  # noqa: F401

from .database import SqlDatabase  # noqa: F401



# -*- coding: utf-8 -*-
"""
lories
~~~~~~


"""

from . import _version

__version__ = _version.get_versions().get("version")
del _version


from .core import (  # noqa: F401
    Constant,
    Directory,
    Directories,
    Configurations,
    ConfigurationError,
    ConfigurationUnavailableError,
    Configurator,
    Resource,
    Resources,
    ResourceError,
    ResourceUnavailableError,
)

from . import data  # noqa: F401
from .data import (  # noqa: F401
    ChannelState,
    Channel,
    Channels,
    Converter,
    ConversionError,
    Listener,
)

from . import connectors  # noqa: F401
from .connectors import (  # noqa: F401
    Connector,
    ConnectorError,
    ConnectionError,
    Database,
    DatabaseException,
    DatabaseUnavailableException,
    Databases,
)

from . import components  # noqa: F401
from .components import (  # noqa: F401
    Component,
    ComponentError,
    ComponentUnavailableError,
)

from . import location  # noqa: F401
from .location import (  # noqa: F401
    Location,
    LocationException,
    LocationUnavailableException,
)

from .components import (  # noqa: F401
    Tariff,
    Weather,
)

from .settings import Settings  # noqa: F401

from . import system  # noqa: F401
from .system import System  # noqa: F401

from . import simulation  # noqa: F401
from .simulation import (  # noqa: F401
    Durations,
    Progress,
    Results,
)

from . import io  # noqa: F401

from . import application  # noqa: F401
from .application import Application  # noqa: F401


def load(name: str = "Lories", **kwargs) -> Application:
    return Application.load(name, **kwargs)



# -*- coding: utf-8 -*-
"""
lories.components.errors
~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from lories._core._component import Component  # noqa
from lories.core.errors import ResourceError, ResourceUnavailableError


class ComponentError(ResourceError):
    """
    Raise if an error occurred accessing the connector.

    """

    # noinspection PyArgumentList
    def __init__(self, component: Component, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.component = component


class ComponentUnavailableError(ResourceUnavailableError, ComponentError):
    """
    Raise if an accessed connector can not be found.

    """



# -*- coding: utf-8 -*-
"""
lories.system
~~~~~~~~~~~~~


"""

from __future__ import annotations

import os
from typing import Any, List, Optional

import pandas as pd
from lories import Configurations
from lories.components import Component, ComponentContext, ComponentUnavailableError, Weather
from lories.core.typing import Timestamp
from lories.location import Location, LocationUnavailableException
from lories.simulation import Results


class System(Component):
    TYPE: str = "system"

    _location: Optional[Location] = None

    @classmethod
    def scan(cls, context: ComponentContext, scan_dir: str, **kwargs) -> List[System]:
        systems = []
        for system_dir in os.scandir(scan_dir):
            if os.path.isdir(system_dir.path):
                kwargs["data_dir"] = system_dir.path
                systems.append(cls.load(context, **kwargs))
        return systems

    # noinspection PyProtectedMember
    @classmethod
    def load(cls, context: ComponentContext, **kwargs) -> System:
        configs = Configurations.load(f"{cls.__name__.lower()}.conf", **kwargs)
        system_id = cls._build_id(context=context, configs=configs)
        if context._contains(system_id):
            context._update(system_id, configs)
            system = context._get(system_id)
        else:
            system = cls(context=context, configs=configs, id=system_id)
            context._add(system)
        return system

    # noinspection PyShadowingBuiltins, PyUnusedLocal
    def __init__(
        self,
        context: ComponentContext = None,
        configs: Optional[Configurations] = None,
        id: Optional[str] = None,
        key: Optional[str] = None,
        name: Optional[str] = None,
        **kwargs,
    ) -> None:
        name = self._build_name(name, configs=configs)
        key = self._build_key(key, configs=configs)
        id = self._build_id(id, key, configs=configs, context=context)
        super().__init__(
            context=context,
            configs=configs,
            id=id,
            key=key,
            name=name,
            **kwargs,
        )

    def configure(self, configs: Configurations) -> None:
        super().configure(configs)
        self.localize(configs.get_member(Location.TYPE, defaults={}))
        self.components.load(configs_dir=self.configs.dirs.conf)

    def localize(self, configs: Configurations) -> None:
        if configs.enabled and all(k in configs for k in ["latitude", "longitude"]):
            self._location = Location(
                configs.get_float("latitude"),
                configs.get_float("longitude"),
                timezone=configs.get("timezone", default="UTC"),
                altitude=configs.get_float("altitude", default=None),
                country=configs.get("country", default=None),
                state=configs.get("state", default=None),
            )
        else:
            self._location = None

    def is_localized(self) -> bool:
        return self._location is not None

    @property
    def location(self) -> Location:
        if self._location is None:
            raise LocationUnavailableException(f"System '{self.name}' has no location configured")
        return self._location

    def has_weather(self) -> bool:
        return self.components.has_type(Weather)

    # noinspection PyTypeChecker
    @property
    def weather(self) -> Weather:
        weather = self.components.get_first(Weather)
        if weather is None:
            raise ComponentUnavailableError(f"System '{self.name}' has no weather configured")
        return weather

    def simulate(
        self,
        start: Timestamp,
        end: Timestamp,
        prior: Optional[pd.DataFrame] = None,
        **kwargs: Any,
    ) -> pd.DataFrame:
        raise NotImplementedError()

    def evaluate(self, results: Results) -> pd.DataFrame:
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
lories.data.listeners.context
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import logging
import time
from collections.abc import Callable
from threading import Lock
from typing import Collection, Optional

import pandas as pd
from lories._core._channel import Channel  # noqa
from lories._core._channels import Channels  # noqa
from lories._core._context import _Context  # noqa
from lories._core._data import DataContext, _DataContext, _DataManager  # noqa
from lories.core import ResourceError
from lories.data.listeners import Listener

# FIXME: Remove this once Python >= 3.9 is a requirement
try:
    from typing import Literal

except ImportError:
    from typing_extensions import Literal


# noinspection PyShadowingBuiltins
class ListenerContext(_Context[Listener]):
    __context: _DataContext
    __lock: Lock

    def __init__(self, context: DataContext, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.__context = self._assert_context(context)
        self._logger = logging.getLogger(self.__module__)

        self.__lock = Lock()

    def __enter__(self) -> ListenerContext:
        self.__lock.acquire()
        return self

    def __exit__(self, type, value, traceback):
        self.__lock.release()

    @classmethod
    def _assert_context(cls, context: DataContext) -> DataContext:
        if context is None or not isinstance(context, _DataManager):
            raise ResourceError(f"Invalid '{cls.__name__}' context: {type(context)}")
        return context

    @property
    def context(self) -> DataContext:
        return self.__context

    # noinspection PyMethodMayBeStatic
    def _create(
        self,
        id: str,
        key: str,
        function: Callable[[pd.DataFrame], None],
        channels: Channels,
        how: Literal["any", "all"] = "any",
        unique: bool = False,
    ) -> Listener:
        return Listener(id, key, function, channels, how, unique)

    # noinspection PyShadowingBuiltins, PyProtectedMember
    def _update(
        self,
        id: str,
        channels: Channels,
        how: Literal["any", "all"] = "any",
        unique: bool = False,
    ) -> None:
        listener = self._get(id)
        if listener._how != how:
            raise ResourceError(f"Trying to register '{how}' updated listener to existing '{listener._how}' instance")
        if listener._unique != unique:
            raise ResourceError(
                f"Trying to register '{unique}' processed listener to existing '{listener._unique}' instance"
            )
        listener.channels.extend(channels)

    # noinspection PyUnresolvedReferences, PyProtectedMember
    def register(
        self,
        function: Callable[[pd.DataFrame], None],
        channels: Channels,
        how: Literal["any", "all"] = "any",
        unique: bool = False,
    ) -> None:
        key = function.__name__
        try:
            context = function.__self__.id
        except AttributeError:
            context = function.__module__
        id = f"{context}.{key}"
        if self._contains(id):
            self._update(id, channels, how, unique)
        else:
            self._add(self._create(id, key, function, channels, how=how, unique=unique))

    def notify(self, *channels: Channel) -> Collection[Listener]:
        listeners = []
        for id, listener in self.items():
            ids = listener.channels.ids
            if any(c.id in ids for c in channels) and listener.has_update():
                if listener.locked():
                    self._logger.warning(
                        f"Listener '{listener.id}' not finished after {round(listener.runtime, 3)} seconds. "
                        f"Please verify your configurations"
                    )
                listeners.append(listener)
        return listeners

    def wait(self, timeout: Optional[float] = None, sleep: Callable = time.sleep) -> None:
        start = time.time()

        def is_timeout() -> bool:
            if timeout is None:
                return False
            return time.time() - start >= timeout

        # noinspection PyShadowingNames
        def has_locked() -> bool:
            locked = [listener.locked() for listener in self.values()]

            if self._logger.getEffectiveLevel() <= logging.DEBUG:
                self._logger.debug(f"Waiting for {len(locked)} listener{'s' if len(locked) > 0 else ''} to finish")
            return any(locked)

        while has_locked() and not is_timeout():
            sleep(0.01)
        if is_timeout():
            pass



# -*- coding: utf-8 -*-
"""
lories.data.validation
~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import datetime as dt

import pandas as pd
from lories._core.typing import Timezone  # noqa
from lories.core.errors import ResourceError  # noqa


def validate_index(data: pd.DataFrame | pd.Series) -> pd.DataFrame | pd.Series:
    if not isinstance(data.index, pd.DatetimeIndex):
        try:
            data.index = pd.to_datetime(data.index)
        except (pd.errors.ParserError, ValueError) as e:
            raise ResourceError(f"Invalid series, without valid DatetimeIndex: {str(e)}")
    if not data.index.is_unique:
        raise ResourceError(f"Invalid series with non unique index: {data[data.index.duplicated()]}")
    return data


def validate_timezone(data: pd.DataFrame | pd.Series, timezone: Timezone) -> pd.DataFrame | pd.Series:
    if isinstance(data, pd.DatetimeIndex):
        data = data.tz_convert(timezone)
    elif pd.api.types.is_datetime64_dtype(data.values):
        data = data.dt.tz_convert(timezone)
    elif data.map(lambda i: isinstance(i, (pd.Timestamp, dt.datetime))).all():
        data = data.map(lambda i: i.astimezone(timezone))
    return data



# -*- coding: utf-8 -*-
"""
lories.components.access
~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

import glob
import os.path
from collections.abc import Callable
from typing import Any, Collection, Mapping, Optional, Sequence, Type

from lories._core._component import Component, _Component, _ComponentContext  # noqa
from lories._core._data import _DataManager  # noqa
from lories.core import Configurations, RegistratorAccess, ResourceError
from lories.util import get_context, update_recursive


# noinspection PyProtectedMember
class ComponentAccess(_ComponentContext, RegistratorAccess[Component]):
    # noinspection PyUnresolvedReferences
    def __init__(self, registrar: Component, **kwargs) -> None:
        context = get_context(registrar, _DataManager).components
        super().__init__(context, registrar, **kwargs)

    # noinspection PyShadowingBuiltins
    def _set(self, id: str, component: Component) -> None:
        if not isinstance(component, _Component):
            raise ResourceError(f"Invalid component type: {type(component)}")

        super()._set(id, component)

    def load(
        self,
        configs: Optional[Configurations] = None,
        configs_file: Optional[str] = None,
        configs_dir: Optional[str] = None,
        configure: bool = False,
        **kwargs: Any,
    ) -> Sequence[Component]:
        return super().load(configs, configs_file, configs_dir, configure, strict=True)

    # noinspection PyShadowingBuiltins, PyUnresolvedReferences
    def load_from_type(
        self,
        cls: Type[Component],
        configs: Configurations,
        type: str,
        key: str,
        name: Optional[str] = None,
        includes: Optional[Collection[str]] = (),
        defaults: Optional[Mapping[str, Any]] = None,
        configure: bool = False,
        sort: bool = True,
        **kwargs,
    ) -> Sequence[Component]:
        kwargs["factory"] = cls
        components = []
        if defaults is None:
            defaults = self._load_registrator_defaults(strict=True)
        if any(i in configs.members for i in includes):
            configs["key"] = key
            configs["name"] = name
        configs = configs.get_member(type, defaults={**configs.get_members(includes), **defaults})
        if any(i in configs.members for i in includes):
            components.append(self._load_from_configs(self._registrar, configs, **kwargs))

        update_recursive(defaults, _Component._build_defaults(configs, includes))

        configs_dirs = configs.dirs.copy()
        configs_members = configs.get_members([s for s in configs.members if s not in defaults])

        components.extend(self._load_from_members(self._registrar, configs_members, defaults=defaults, **kwargs))

        if "alias" in configs:
            key = configs.get("alias")
        for configs_path in glob.glob(str(configs_dirs.conf.joinpath(f"{key}*.conf"))):
            if configs.name == os.path.basename(configs_path):
                continue

            configs = Configurations.load(
                configs_path,
                **configs_dirs.to_dict(),
                **defaults,
            )
            components.append(self._load_from_configs(self._registrar, configs, **kwargs))

        if sort:
            self.sort()
        if configure:
            self.configure(components)
        return components

    def _create(
        self,
        context: _ComponentContext | _Component,
        configs: Configurations,
        factory: Optional[Callable[..., Component]] = None,
        **kwargs: Any,
    ) -> Component:
        if factory is None:
            factory = super()._create
        return factory(context, configs, **kwargs)



# -*- coding: utf-8 -*-
"""
lories.application.view.pages.components.page
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from collections.abc import Sequence
from typing import Collection, Generic, Optional

import dash_bootstrap_components as dbc
from dash import Input, Output, callback, html

import pandas as pd
from lories.application.view.pages import Page, PageLayout
from lories.typing import Channel, Channels, Component, Components, Configurations, Connectors, Data


class ComponentPage(Page, Generic[Component]):
    _component: Component

    def __init__(self, component: Component, *args, **kwargs) -> None:
        super().__init__(
            id=component.id,
            key=component.key,
            name=component.name,
            *args,
            **kwargs,
        )
        self._component = component

    @property
    def configs(self) -> Configurations:
        return self._component.configs

    @property
    def connectors(self) -> Connectors:
        return self._component.connectors

    @property
    def components(self) -> Components:
        return self._component.components

    @property
    def data(self) -> Data:
        return self._component.data

    def is_active(self) -> bool:
        return self._component.is_active()

    def create_layout(self, layout: PageLayout) -> None:
        layout.card.add_title(self.title)
        layout.card.add_footer(href=self.path)
        layout.append(html.H4(f"{self.title}:"))

    def _on_create_layout(self, layout: PageLayout) -> None:
        super()._on_create_layout(layout)
        self._create_data_layout(layout, self.data.channels)

    def _create_data_layout(self, layout: PageLayout, channels: Channels, title: Optional[str] = "Data") -> None:
        if len(self.data.channels) > 0:
            layout.append(html.Hr())

            data = []
            if title is not None:
                data.append(html.H5(f"{title}:"))
            data.append(self._build_data(channels))

            layout.append(dbc.Row(data))

        # TODO: append data-update separately to view

    def _build_data(self, channels: Channels) -> html.Div:
        @callback(
            Output(f"{self.id}-data", "children"),
            Input("view-update", "n_intervals"),
        )
        def _update_data(*_) -> Sequence[dbc.AccordionItem]:
            return [self._build_channel(channel) for channel in channels]

        return html.Div(
            [
                dbc.Accordion(
                    id=f"{self.id}-data",
                    children=_update_data(),
                    start_collapsed=True,
                    always_open=True,
                    flush=True,
                ),
            ]
        )

    def _build_channel(self, channel: Channel) -> dbc.AccordionItem:
        return dbc.AccordionItem(
            title=dbc.Row(
                [
                    dbc.Col(self._build_channel_title(channel), width="auto"),
                    dbc.Col(self._build_channel_header(channel), width="auto"),
                ],
                justify="between",
                className="w-100",
            ),
            children=[
                dbc.Row(
                    [
                        dbc.Col(html.Span("Updated:", className="text-muted"), width=1),
                        dbc.Col(self._build_channel_timestamp(channel), width="auto"),
                    ],
                    justify="start",
                ),
                dbc.Row(
                    [
                        dbc.Col(None, width=1),
                        dbc.Col(self._build_channel_body(channel), width="auto"),
                    ],
                    justify="start",
                ),
            ],
            id=f"{self.id}-data-{self._encode_id(channel.key)}",
        )

    # noinspection PyMethodMayBeStatic
    def _build_channel_title(self, channel: Channel) -> html.Span:
        # TODO: Implement future improvements like the separation of name and unit
        return html.Span(channel.name, className="mb-1")

    # noinspection PyMethodMayBeStatic
    def _build_channel_header(self, channel: Channel) -> Collection[html.Span]:
        channel_header = []
        if channel.is_valid():
            channel_header.append(self._build_channel_value(channel))
            channel_header.append(self._build_channel_unit(channel))
        channel_header.append(self._build_channel_state(channel))
        return channel_header

    # noinspection PyMethodMayBeStatic
    def _build_channel_body(self, channel: Channel) -> Optional[html.Div]:
        if not channel.is_valid() or not (channel.has_logger() or channel.type == pd.Series):
            return None

        # TODO: Implement a Graph for logged values or pandas.Series types
        return html.Div(html.I("Placeholder", className="text-muted"))

    # noinspection PyMethodMayBeStatic
    def _build_channel_timestamp(self, channel: Channel) -> html.Small:
        timestamp = channel.timestamp
        if not pd.isna(timestamp):
            timestamp = timestamp.isoformat(sep=" ", timespec="seconds")
        return html.Small(timestamp, className="text-muted")

    # noinspection PyMethodMayBeStatic
    def _build_channel_value(self, channel: Channel) -> html.Span:
        # TODO: Implement further type validation
        value = channel.value
        if not pd.isna(value):
            if channel.type == float:
                value = round(channel.value, 2)
        return html.Span(value, className="mb-1", style={"margin-right": "0.2rem"})

    # noinspection PyMethodMayBeStatic
    def _build_channel_unit(self, channel: Channel) -> html.Span:
        return html.Span(channel.unit, className="text-muted", style={"margin-right": "2rem"})

    # noinspection PyMethodMayBeStatic
    def _build_channel_state(self, channel: Channel) -> html.Small:
        state = str(channel.state).replace("_", " ")
        color = "success" if channel.is_valid() else "warning"
        if state.lower().endswith("error") or state.lower() == "disabled":
            color = "danger"
        return html.Small(state.title(), className=f"text-{color}", style={"margin-right": "1rem"})



# -*- coding: utf-8 -*-
"""
lories.data.converters.io.voltage
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from lories.core.typing import Configurations
from lories.data.converters import register_converter_type
from lories.data.converters.io.analog import AnalogInput


# noinspection PyAbstractClass
@register_converter_type("voltage_input", "voltage_sensor")
class VoltageInput(AnalogInput):
    INPUT_KEY: str = "voltage"

    def _configure_input(
        self,
        configs: Configurations,
        default_max: float = None,
        default_min: float = 0.0,
    ) -> None:
        default_max = configs.get_float("voltage", default=default_max)
        super()._configure_input(configs, default_max, default_min)



# -*- coding: utf-8 -*-
"""
lories._core._resources
~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Generic, Iterable, Iterator, Sequence, Tuple, TypeVar, Union, overload

from lories._core._resource import Resource, _Resource


class _Resources(ABC, Generic[Resource], Sequence[Resource]):
    @abstractmethod
    def append(self, resource: Resource) -> None: ...

    @abstractmethod
    def extend(self, resources: Iterable[Resource]) -> None: ...

    @abstractmethod
    def update(self, resources: Iterable[Resource]) -> None: ...

    @abstractmethod
    def ids(self) -> Sequence[str]: ...

    @abstractmethod
    def keys(self) -> Sequence[str]: ...

    @abstractmethod
    def copy(self) -> Resources: ...

    @abstractmethod
    def apply(self, apply: Callable[[Resource], Resource], inplace: bool = False): ...

    # noinspection PyShadowingBuiltins
    @overload
    def filter(self, filter: Callable[[Resource], bool]): ...

    @abstractmethod
    def filter(self, *filters: Callable[[Resource], bool]): ...

    # noinspection PyShadowingBuiltins, SpellCheckingInspection
    @abstractmethod
    def groupby(self, by: Callable[[Resource], Any] | str) -> Iterator[Tuple[Any, Resources]]: ...

    @abstractmethod
    def to_configs(self) -> Dict[str, Any]: ...


Resources = TypeVar("Resources", bound=_Resources[_Resource])

ResourcesArgument = TypeVar(
    "ResourcesArgument",
    bound=Union[_Resource, _Resources[_Resource], Iterable[_Resource], Iterable[str], str],
)



# -*- coding: utf-8 -*-
"""
lories.application.view.login.user
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations

from flask_bcrypt import check_password_hash, generate_password_hash
from flask_login import UserMixin


# noinspection PyShadowingBuiltins
class User(UserMixin):
    name: str
    hash: str

    _authenticated: bool = False

    @classmethod
    def from_password(cls, name: str, password: str) -> User:
        return cls(name, generate_password_hash(password))

    @property
    def id(self) -> str:
        return self.name

    def __init__(self, name: str, hash: str) -> None:
        self.name = name
        self.hash = hash

        self._authenticated = False

    # noinspection PyShadowingBuiltins
    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.name}:{self.hash[8:]}*)"

    # noinspection PyShadowingBuiltins
    def __str__(self) -> str:
        vars = [
            f"name = {self.name}",
            f"hash = {self.hash[8:]}*",
        ]
        return f"{type(self).__name__}:\n\t" + "\n\t".join(vars)

    @property
    def is_authenticated(self) -> bool:
        return self._authenticated

    def authenticate(self, password: str) -> bool:
        self._authenticated = check_password_hash(self.hash, password)
        return self._authenticated
